%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           ibmrs-asm.SL
% Title:          IBM RS 6000 specific information for LAP-TO-ASM
% Author:         Winfried Neun, ZIB Berlin
% Created:        November 5, 1990
% Package:        Compiler
%
% (c) Copyright 1990, Konrad-Zuse Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn (load if!-system common)
        (off usermode)))

(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
         outputsymfile!* commentformat!* labelformat!* 
         externaldeclarationformat!* exporteddeclarationformat!* 
         fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
         reservedatablockformat!* definedfunctioncellformat!* 
         undefinedfunctioncellinstructions!* mainentrypointname!* 
         codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
         numericregisternames!* pathin!*))
%%
%%      DG assembler must declare global name before use.
%%
(fluid '(*declarebeforeuse))

(setq *declarebeforeuse t)

(setq codefilenameformat!* "%w.s")

(setq datafilenameformat!* "d%w.s")

(setq inputsymfile!* "Ibmrs.sym")

(setq outputsymfile!* "Ibmrs.sym")

(setq mainentrypointname!* (intern "_psl_main"))

(setq commentformat!* "#  %p %n")

(setq labelformat!* "%w:")

(setq exporteddeclarationformat!* "	.globl	%w%n")
(setq externaldeclarationformat!* "	.globl	%w%n")

(setq fullwordformat!* "	.long	%e%n")

% FullWord expects %e for parameter                                        

(setq doublefloatformat!* "	double	0d%w%n")

(setq reservezeroblockformat!* "	.comm	%w,%e%n")

(setq reservedatablockformat!* "	.space	%e%n")

(setq definedfunctioncellformat!* "	.long %w%n")

(setq undefinedfunctioncellinstructions!*
      '((!.long undefinedfunction)))

%% From lap-to-asm: 

(de initsymfnc1 (x)
  (prog (ep)
        (setq ep (get x 'entrypoint))
        (if (null ep)
          (dataprintundefinedfunctioncell (id2int x))
          (dataprintdefinedfunctioncell ep x))))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
        (setq oldout (wrs dataout*))
        (dataprintf "	.long  undefinedfunction%n")
        (wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
   (datadeclareexternal x)
     % dont disturb undefined function and interpreter
   (dataprintf definedfunctioncellformat* x (id2int idn)))

(fluid '(&links-present& locallabels*))

(de asmoutlap (u)
  (prog (locallabels* oldout)
        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
                                  (assoc '*call u)))

        (setq u (pass1lap (&fillframeholes u)))
        (setq u (lapoptreformjumps u)) % lapopt pass
        % Expand cmacros, quoted expressions
        (codeblockheader)
        (setq oldout (wrs codeout*))
        (foreach x in u do (asmoutlap1 x))
        (wrs oldout)
        (codeblocktrailer)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(deflist '((r0 0) (r30 30) (r29 29) (r28 28) (r27 27) (r26 26) (r25 25) 
   (r24 24) (r 23 23) (r22 22)
   (true 12) (false 4) (eq 2) (lt 0) (gt 1) (always 20)) 'symbol)

(setq asmopenparen!* '!()

(setq asmcloseparen!* '!))

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
         'binaryasmop)

(put 'lnot 'unaryasmop '!!)

%(put 'mkitem 'asmexpressionformat "(%e<<27)+%e")
(fluid '(*mkitemformat))
(setq *mkitemformat "%e") % ZIB

(put 'mkitem 'asmexpressionfunction 'asm-mkitem)

(de asm-mkitem (mkitem-tag-inf)
  (printf
    *mkitemformat
    `(printexpression ,(caddr mkitem-tag-inf))
    `(printexpression ,(cadr mkitem-tag-inf))))

(put 'bitfield 'operandprintfunction 'asm-bitfield)

(de asm-bitfield (w)
     (prin2 (eval (cadr w))) (prin2 "<") (prin2 (eval (caddr w))) (prin2 ">"))

(global '(*writingasmfile))

(de codefileheader ()

  (setq *writingasmfile t)      %% Somebody must set this flag. but who?

  (codeprintf "     .set   r0,0%n")
  (codeprintf "     .set   r1,1%n")
  (codeprintf "     .set   r2,2%n")
  (codeprintf "     .set   r3,3%n")
  (codeprintf "     .set   r4,4%n")
  (codeprintf "     .set   r5,5%n")
  (codeprintf "     .set   r6,6%n")
  (codeprintf "     .set   r7,7%n")
  (codeprintf "     .set   r8,8%n")
  (codeprintf "     .set   r9,9%n")
  (codeprintf "     .set   r10,10%n")
  (codeprintf "     .set   r11,11%n")
  (codeprintf "     .set   r12,12%n")
  (codeprintf "     .set   r13,13%n")
  (codeprintf "     .set   r14,14%n")
  (codeprintf "     .set   r15,15%n")
  (codeprintf "     .set   r16,16%n")
  (codeprintf "     .set   r17,17%n")
  (codeprintf "     .set   r18,18%n")
  (codeprintf "     .set   r19,19%n")
  (codeprintf "     .set   r20,20%n")
  (codeprintf "     .set   r21,21%n")
  (codeprintf "     .set   r22,22%n")
  (codeprintf "     .set   r23,23%n")
  (codeprintf "     .set   r24,24%n")
  (codeprintf "     .set   r25,25%n")
  (codeprintf "     .set   r26,26%n")
  (codeprintf "     .set   r27,27%n")
  (codeprintf "     .set   r28,28%n")
  (codeprintf "     .set   r29,29%n")
  (codeprintf "     .set   r30,30%n")
  (codeprintf "     .set   r31,31%n")
  (codeprintf "     .globl pslmain.s[rw]%n")
  (codeprintf "     .globl SYMFNC[rw]%n")
  (codeprintf "     .globl SYMVAL[rw]%n")
  (codeprintf "     .toc%n")
  (codeprintf "T.pslmain.s:	.tc	pslmain.s[tc],pslmain.s[rw]%n")
  (codeprintf "T.SYMFNC:       .tc     T.symfnc[tc],SYMFNC[rw]%n")
  (codeprintf "T.symval:       .tc     T.symval[tc],SYMVAL[rw]%n")
  (codeprintf "T.pslmainc:     .tc     T.pslmainc[tc],pslmain[PR]%n")
  (codeprintf "     .globl  pslmain[ds]%n")
  (codeprintf "     .long   pslmain[PR]%n")
  (codeprintf "     .long   TOC[tc0]%n")
  (codeprintf "     .long 0%n")
  (codeprintf "     .toc%n")
  (codeprintf "T.pslmain: .tc  .pslmain[tc],pslmain[ds]%n")
  (codeprintf "     .globl pslmain[PR]%n")
  (codeprintf "     .csect pslmain[PR]%n"))

(de datafileheader ()
 nil)

(de codefiletrailer ()
  nil)

(de datafiletrailer ()
  nil)

(de codeblockheader ()
  (codeprintf "	.align	4%n"))

(de codeblocktrailer ()
  nil)

(de dataalignfullword ()
  (dataprintf "	.align	4%n"))

(de printstring (s)
  (prog (n)
        (setq n (size s))
        (for (from i 0 n 1) (do (printbyte (indx s i))))
        (printbyte 0)
        (printf "	.align	4%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf "	.byte ")
  (printexpression x)
  (terpri))

(de printhalfwordlist (l)
  (foreach x in l do (printhalfword x)))

(de printhalfword (x)
  (printf "     half ")
  (printexpression x)
  (terpri))

(de printopcode (x)
  (prog (!*lower)
        (setq !*lower t)
        (prin2 "     ")(prin2 x)))

(de specialactionformainentrypoint ()
  (codeprintf "	.globl	_psl_main%n_psl_main:"))

(de asmsymbolp (x)
  (unixsymbolp (if (idp x)
                 (id2string x)
                 x)))

(de unixsymbolp (x)
  (prog (n c i)
        (setq n (size x))
        (setq c (indx x 0))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) (equal c (char !.)) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i 1)
   loop
        (when (greaterp i n)
          (return t))
        (setq c (indx x i))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (equal c (char _)) 
                    (and (geq c (char !0)) (leq c (char !9))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i (plus i 1))
        (go loop)))

(de printnumericoperand (x)
  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (ibmrs-register-p x)
    (progn (printoperand x))
    (error 0 (list "In MCibmrs, deffered operand can't be used like this:" x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))
  (printexpression (cadr x))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!) ))

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintindexed (x)
  (setq x (cdr x))
  (printexpression (cadr x))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!) ))

(put 'indexed 'operandprintfunction 'operandprintindexed)

(de operandprintscaled (x)
  (setq x (cdr x))
  (printoperand (car x))
  (prin2 '![ )
  (printoperand (cadr x))
  (prin2 '!] ))

(put 'scaled 'operandprintfunction 'operandprintscaled)

(de operandprintimmediate (x)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)
%%
%%      high and low is special keyword for ibmrs-aviion assembler.
%%
(de operandprinthigh (x)
  (setq x (cadr x))
  (when (not (numberp x)) (errorprintf " *** wrong immediate : %w %n" x))
  (prin2 (wshift x -16)))

(put 'high 'operandprintfunction 'operandprinthigh)
(put 'hi16 'operandprintfunction 'operandprinthigh)

(de operandprintlow (x)
  (setq x (cadr x))
  (when (not (numberp x)) (errorprintf " *** wrong immediate : %w %n" x))
  (prin2 (wand x 16#ffff)))


(put 'low 'operandprintfunction 'operandprintlow)
(put 'lo16 'operandprintfunction 'operandprintlow)

(de operandprintwarray (x)
  (prin2 (cadr x)))

(put 'WARRAY 'operandprintfunction 'operandprintwarray)

(de operandprintforeignentry (x)
  (prog (!*lower symnamx)
        (setq symnamx (symnam (id2int (cadr x))))
        (setq !*lower t)
        (when (not (eq (id2int '!.)(strbyt (strinf symnamx) 0)))
           (prin2 ".")) % dont change .mul into _.mul
        (prin2 (cadr x)) (prin2 "[PR]")))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf "# %p %n" x)
  )

(de instructionprint (x)
  (codeprintf "#  %p %n" x)
  )

(de !*cerror (x)
  (prog (i)
        (setq i (wrs nil))
        (printf "%n *** CERROR: %r %n " x)
        (wrs i)
        (return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)
%%
%%      jmpl has something different syntax in sun4 `as'.
%%
(put 'jmpl 'asmpseudoop 'printjmpl)

(de printjmpl (x)
  (let ((src (cadr x))
        (dst (caddr x)))
       (prin2 " jmpl ")
       (cond ((eq (car src) 'deferred)
              (printoperand (cadr src)))
             ((eq (car src) 'displacement)
              (printoperand (cadr src))
              (print-secondary-source (caddr src)))
             (t
              (printoperand (car src))
              (cond ((cdr src)
                     (print-secondary-source (cadr src))))))
       (prin2 ",")
       (printoperand dst)
       (terpri)))
%%
%%      (word label)
%%
(put 'word 'asmpseudoop 'print-word)

(de print-word (x)
  (prin2 "	.long ")
  (printoperand (cadr x))
  (terpri))

(put '*entry 'asmpseudoop 'asmprintentry)
(fluid '(*mainfound))
(de asmprintentry (x)
  (prog (y)
        (printcomment x)
        (setq x (cadr x))
        (setq y (findentrypoint x))
        (unless (flagp x 'internalfunction)
          (findidnumber x))
        (if (eq x mainentrypointname*)
          (progn (setq *mainfound t)
                 (specialactionformainentrypoint))
          (codedeclareexporteduse y))))

(de codedeclareexporteduse (y)
    (codedeclareexported y)
    (printlabel y)
    (terpri))

(de findentrypoint (x)
  (cond ((get x 'entrypoint) (get x 'entrypoint))
        ((and (asmsymbolp x)
              (not (get x 'symbol))
              (not (flagp x 'foreignfunction)))
         (put x 'entrypoint x) 
         x)
        (t 
         (let ((name (stringgensym)))
           (put x 'entrypoint name)
           name))
        ))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = "))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = ")
  (prin2 (caddr x))
  (terpri))

(put 'define 'asmpseudoop 'printdefine)

(de print-register-or-immediate(name)
  (cond ((ibmrs-register-p name) (printoperand name))
        (t (printoperand name))))

(de print-immediate (val)
  (cond ((atom val)
         (printoperand val))
        ((eq (car val) 'WARRAY)
         (prin2 (cadr val)))
        (t
         (error 0 (list "Unknown constant expression:" val)))))

(de print-register-name(name)
  (cond ((numberp name) (setq *lower t) (prin2 (to-standard-name name))
					(setq *lower nil))
        ((get name '*register-code) (setq *lower t)
                                    (prin2 (to-standard-name name ))
                                    (setq *lower nil))
        ((get name '*f-register-code)
         (prin2 "%f")
         (prin2 (get name '*f-register-code)))
        ((get name '*c-register-code)
         (prin2 "%c")
         (prin2 (get name '*c-register-code)))
        (t
         (printoperand name))))         %% if reg-name is not a register, it
                                        %% must be a constant.

%%
%%
%%      This function is originally defined in lap-to-asm.sl.
%%
(de printoperand (x)
  (setq x (to-standard-name x))
  (cond ((stringp x) (prin2 x))
        ((numberp x) (printnumericoperand x))
        ((ibmrs-register-p x) (print-register-name x))
        ((idp x) (prin2 (findlabel x)))
%       ((ibmrs-register-p (car x)) (print-src-operand x))
        (t (prog (hd fn)
                 (setq hd (car x))
                 (cond ((setq fn (get hd 'operandprintfunction)) 
                        (apply fn (list x)))
                       ((and (setq fn (getd hd)) (equal (car fn) 'macro)) 
                        (printoperand (apply (cdr fn) (list x))))
                       ((setq fn (wconstevaluable x)) (printoperand fn))
                       (t (printexpression x)))))))
%
%       print operand by [rs1+rs2] format.
%
(de print-src-operand (src)
  (printoperand (car src))
  (and (cdr src)
       (print-secondary-source (cadr src)))
  )

(de print-secondary-source (src)
  (cond ((and (pairp src) (eq (car src) '-))
         (cond ((numberp (cadr src))
                (cond ((minusp (cadr src)) (prin2 ",")
                       			   (prin2 (minus (cadr src))))
                      (t (prin2 "-") (prin2 (cadr src)))))
               (t (prin2 "-") (printoperand (cadr src)))))

        ((numberp src)
         (cond ((minusp src) (prin2 "-") (prin2 (minus src)))
               (t (prin2 ",") (printoperand src))))
        (t (prin2 ",") (printoperand src))))

% we have to put the true mkitems (namely the strings in the kernel)
% to data object. SIgh

(fluid '(stringlabels*))

(de AppendOneConstant (ExpressionLabelPair)
  (if (stringp (car ExpressionLabelPair))
      (prog (expa)
	(setq expa (ExpandItem (car ExpressionLabelPair)))
 (datadeclareexternal (caddr expa))
 (dataprintf "%w:   .long %w %n" (cdr ExpressionLabelPair) (caddr expa))
	(codedeclareexternal (caddr expa))
        (codedeclareexternal (cdr ExpressionLabelPair))
	(push (cons (cdr ExpressionLabelPair) (caddr expa))
		stringlabels!*)
        (return (appenditem 0 )))
  (progn (AddCodeLabel (cdr ExpressionLabelPair))
         (AppendItem (car ExpressionLabelPair)))))

(de writestringmkitemword()
  NIL) (commentoutcode
        (codeprintf "    .globl   hilfe[rw]%n")
	(codeprintf "    .csect   hilfe[rw]%n")
	(mapc stringlabels!* (function (lambda (x)
		(codeprintf "%w:   .long %w %n" (car x) (cdr x))))))
