%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           $pxc/spectrum-lap.sl
% Title:          HP RISC PSL assembler
% Author:         Winfried Neun,
% Created:        November 15, 1991
% Modified:
% Status:         Experimental
%
% Copyright (c) 1991, Konrad-Zuse-Zentrum Berlin, All Rights Reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Revisions:
% Revisions
%
%  22.July 1993 (Winfried Neun)
% Fixed ... Added reformation for long conditional jumps
%  25. Jan 1993 (Winfried Neun)
% Added reformation for long conditional jumps
%

(compiletime (load fasl-decls common))

(global '(*writingasmfile))

(fluid '(labeloffsets* currentoffset* codesize* codebase* entries*
	 forwardinternalreferences* lapreturnvalue* *writingfaslfile
	 initoffset* *pgwd *pwrds label-alist* *usefastframe))

(compiletime (ds labelp (x) (atom x)) )

(fluid '(&links-present& *usesafecondjumps))

(de lap (u)
  (prog (labeloffsets* lapreturnvalue* entries* *usesafecondjumps
	*usefastframe)
	(unless *writingfaslfile
	  (setq currentoffset* 0)
	  (setq codebase* (gtbps 0)))

	(setf &links-present& (or (look-for-links u)
		                  (assoc '*Id-apply u)
		                  (assoc '*call u)))

%       (when (greaterp (length u) 2048) (setq *usesafecondjumps t))
	(setq u (pass1lap (&fillframeholes u)))
	(setq u (lapoptreformjumps u))
	(setq u (reformcondjumps u))
    
	(when *pgwd
	  (foreach x in u do (if (atom x)
		     (prin2 x)
		     (printf "          %p%n" x))))

	%
	%       findlabels calculates codesize and fixup label address
	%
	(findlabels u)

	(unless *writingfaslfile
		(unless (eq nextbps (wand nextbps (wnot 7))) (gtbps 1))
		(setq codebase* (gtbps (wquotient (wplus2 codesize!* 3) 4))))

	%
	%       generate machine code
	%
	(foreach x in u do (depositinstruction x))

	(when (not *writingfaslfile) (flushcache codebase* codesize*))

	(defineentries)

	(when (and (not *writingfaslfile) *pwrds)
	  (errorprintf "*** %p: base 16#%x, length %d bytes" 
		       (foreach x in entries* collect (first (car x))) 
		       codebase* codesize*))
	(return (if lapreturnvalue* (mkitem code-tag lapreturnvalue*) NIL))))

(de look-for-links (u)
       (cond ((null u) nil)
	     ((atom (car u)) (look-for-links (cdr u)))
	     ((eq (caar u) '*linke)
		  (cond ((flagp (caddar u) 'foreignfunction) t)
		         (t (look-for-links (cdr u)))))
	     ((eq (caar u) '*link)
		  (cond
		    ((eq  (cadar u) 'catchsetup) t) % is opencode
		         % but problems will arise in throw if no
		         % allocation takes place
		    ((memq (cadar u) '(fast-codeapply fast-idapply)) t)
		    ((get (cadar u) 'opencode) (look-for-links (cdr u)))
		    (t t)))
	     (t (look-for-links (cdr u)))
)      )

(de findlabels (u)
     (setq codesize* 0)
     (setq label-alist* nil)
     (foreach x in u do (setq codesize* (plus2 (onelaplength x) codesize*))))

(de reformcondjumps(u)
        % conditional jumps have to be reformed eventually.
        % if they are going "far away"
        % There is an 8k byte max displacement. Sigh
	% 
	%  cond_jump  cond arg1 arg2 target
	%  delay_inst                           =>
	%
	%  jump (inverted cond) arg1 arg2 lab
	%  nop
	%  bl target
	%lab
	%  delay_inst

     (if (wlessp (length u) 2000) u
        (prog (uu uuu pos lab newinst)
           (setq pos 0)
           (setq uuu u)
        loop
           (setq uu (car u))
           (cond ((atom uu) nil)
                ((and (memq (car uu) '(comb comib))
                      (wgreaterp
                        (abs
			(wdifference 
				(wdifference (length uuu)
				 (length (member (cadr (cdddr uu)) uuu)))
                                     pos)) 1000))

			(setq lab (intern (gensym)))
			(setq newinst 
				(cons '(nop)
				    (cons `(bl ,(cadr (cdddr uu)) (reg 0))
					(cons lab 
						(cdr u)))))
			(rplacd u newinst)
			(setq newinst (list (car uu)
				(list 'compl (cdr (assoc (cadr (cadr uu))
				   '((!= . !<!>) (!< . !>!=) (!> . !<!=)
				     (!<!> . !=) (!>!= . !<) (!<!= . !>)))))
				(caddr uu) (cadddr uu) lab))
			(rplaca u newinst)))
           (setq pos (wplus2 pos 1))
           (setq u (cdr u))
           (when (null u) (return uuu))
           (go loop))))


(de onelaplength (x)  
   (cond ((atom x) (push (cons x (plus2 currentoffset* codesize*))
		         label-alist*) 0)
	 ((eq (car x) 'string) (wtimes2 addressingunitsperitem 
				(strpack (size (second x))) ))
	 ((eq (car x) 'oddalign)
		 (if (eq (plus2 currentoffset* codesize*) 
			(wand (plus2 currentoffset* codesize*) (wnot 7))) 
					addressingunitsperitem 0))
	 ((eq (car x) '*entry) (setq x (second x))
		(push (cons x (plus2 currentoffset* codesize*)) label-alist*) 
		            0)
	 ((get (first x) '*op-length*) (get (first x) '*op-length*))
	 (t (error 4711 (list " No length for" x)))))

(de saveentry (x)
   (cond ((equal (second x) '***code**pointer***) 
	 (setq lapreturnvalue* 
		(if *writingfaslfile currentoffset*
		             (wplus2 codebase* currentoffset*))))
	((not *writingfaslfile)
	  (setq entries* (cons (cons (rest x) currentoffset*) entries*))
	  (unless lapreturnvalue*
	    (setq lapreturnvalue* (wplus2 codebase* currentoffset*))))
	((equal (second x) '**fasl**initcode**) 
	  (setq initoffset* currentoffset*))
	((flagp (second x) 'internalfunction) 
	  (put (second x) 'internalentryoffset currentoffset*))
	(t (findidnumber (second x)) 
	   (dfprintfasl (list 'putentry (mkquote (second x)) 
		         (mkquote (third x)) currentoffset*)))))

(de defineentries ()
  (foreach x in entries* do 
	   (putd (first (car x)) (second (car x)) 
		 (mkcode (wplus2 codebase* (cdr x))))))

(de labeloffset (x)
  (cond ((codep x) (if *writingfaslfile  (inf x)
                         (wdifference (inf x) codebase*)))
	((setq x (assoc x label-alist*)) (cdr x))
	(t (stderror (bldmsg "Internal entry forward reference: %r" x)))))

(de depositfloat (x)     % this will not work in cross-assembly                                   
  (progn (setq x (fltinf (second x)))
	 (depositword (floathighorder x))
	 (depositword (floatloworder x))))

(de depositbyte (x)
  (setf (byte (wplus2 codebase* currentoffset*) 0) x)
  (updatebittable 1 0)
  )

(de depositword (x)
  (putmem (wplus2 codebase* currentoffset*) x)
  (updatebittable 4 0)
  (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem)))

(de deposit-relocated-word (offset)
  % Given an OFFSET from CODEBASE*, deposit a word containing the
  % absolute address of that offset.
  (putmem (wplus2 codebase* currentoffset*)
	  (wplus2 offset (if *writingfaslfile 0 codebase*)))
  (updatebittable 4 reloc-word)
  (setq currentoffset* (wplus2 currentoffset* addressingunitsperitem)))
  
(de depositwordexpression (x)  % Only limited expressions now handled                                   
  (let (y)
  (cond ((fixp x) (depositword (int2sys x)))
	((stringp x) (deposit-relocated-word 
                     (if *writingfaslfile (labeloffset x)
				(labeloffset x) )))
	((memq  (first x) '($fluid $global fluid global))
		        (depositvaluecelllocation (second x)))
	((equal (first x) 'mkitem) (deposititem (second x) (third x)))
	((equal (first x) 'foreignentry)
		                   (depositfunctioncelllocation (second x)))
        ((equal (first x) 'immediate)
         (cond ((and (pairp (second x))
                     (equal (first (second x)) 'mkitem))
                     (depositwordexpression (second x)))
               ((setq y (assoc (second x) label-alist*))
                (deposit-relocated-word (cdr y)))
               ((and (pairp (second x))
                     (equal (first (second x)) 'entry))
                (depositfunctioncelllocation (second (second x))))
               ((and (pairp (second x)))
                     (equal (first (second x)) 'idloc)
                (depositwordidnumber (second (second x))))

	       (t (stderror (bldmsg "Undefined label: %w" (second x))))))

	((setq y (wconstevaluable x))    (depositword (int2sys y)))
	(t  (stderror (bldmsg "Expression too complicated %r" x))))))

(de depositwordidnumber (x) 
  (cond
    ((or (not *writingfaslfile) (wleq (idinf x) 128)) (depositword (idinf X)))
    (t (putword (wplus2 codebase* currentoffset*) 0 
	       (makerelocword reloc-id-number (findidnumber x))) 
       (updatebittable 4 reloc-word))))

(de deposititem (tagpart infpart)
  (if (not *writingfaslfile)
    (depositword 
       (mkitem tagpart 
	  (cond ((labelp infpart) (wplus2 codebase* (labeloffset infpart)))
		((equal (first infpart) 'idloc)    (idinf (second infpart)))
		(t (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))))
    (progn 
	(cond ((labelp infpart) (putmem (wplus2 codebase* currentoffset*) 
		                                        % RELOC_CODE_OFFSET = 0
		                      (mkitem tagpart (labeloffset infpart))))
	      ((equal (first infpart) 'idloc) 
		 (putmem (wplus2 codebase* currentoffset*)
		         (mkitem tagpart 
		          (makerelocinf reloc-id-number
		           (findidnumber (second infpart))))))
	      (t  (stderror (bldmsg "Unknown inf in MkItem %r" infpart))))
	(updatebittable 4 reloc-inf))
        (setq currentoffset* (iplus2 addressingunitsperitem currentoffset*))))



(de systemfaslfixup ()
  (prog (x type label place fwdir w1 w2 w)
	(while forwardinternalreferences*
	  (setq fwdir (car forwardinternalreferences*))
	  (setq type (car fwdir) label (cadr fwdir) place (caddr fwdir))
	  (if (null (get label 'internalentryoffset))
	    (errorprintf "*** %r not defined in this module, call is incorrect"
		         (cadr fwdir))
	    (cond ((eq type 'CALL) % we use BL instr, have to prepare the w's
		   (setq w (wdifference (get label 'internalentryoffset) place))
		   (setq w (assemble_17 (wshift (wdifference w 8) -2)))
                   (setq w1 (cdr (assoc 'w1 w)))
		   (setq w2 (cdr (assoc 'w2 w)))
		   (setq w (cdr (assoc 'w w)))
 		   (putmem (wplus2 codebase* place)
		     (wor (wshift (wshift (getmem (wplus2 codebase* place))
					-21) 21)
		      (wor (wshift w1 16)
			  (wor (wshift w2 2) w)))))
		  (t  (error 0 (list "Strange forward internal reference: "
		                  fwdir)))))
	  (setq forwardinternalreferences* (cdr forwardinternalreferences*)))
	(mapobl (function (lambda (y) (remprop y 'internalentryoffset))))
))

(de makerelocword (reloctag relocinf)
  (wplus2 (wshift reloctag 30) (field relocinf 2 30)))

(de makerelocinf (reloctag relocinf)
  (wplus2 (wshift reloctag 22) (field relocinf 10 22)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(define-constant reloc-movex 1)

(de deposit-relocatable-movex (code)
		                 % code is (add r,0,r) where r is the dst.
  %(setf (byte (wplus2 codebase* currentoffset*) 0) (wshift code -24))
  %(setf (byte (wplus2 codebase* currentoffset*) 1) 
%	(wand 16#ff (wshift code -16)))
%  (setf (byte (wplus2 codebase* currentoffset*) 2)
%	(wand 16#ff (wshift code -8)))
%  (setf (byte (wplus2 codebase* currentoffset*) 3)
%	(wand 16#ff code))
  (putmem (wplus2 codebase* currentoffset*) code)
  (updatebittable 1 reloc-word)
  (updatebittable 3 reloc-movex) )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(define-constant reloc-xidloc 3)

(de addressvalue (label)
  (cond (*writingfaslfile (fasladdressvalue label))
	(t (lapaddressvalue label))))

(de fasladdressvalue (label)
  (cond ((numberp label) label)
	((assoc label label-alist*)   % we must not relocate these for MIPS
	        (cdr (assoc label label-alist*)) )
	((atom label)  (error 0 (list "Undefined label:" label)))
        ((eq (car label) 'plus) (fasladdressvalue (caddr label)))
	((eq (car label) 'idloc)
		(updatebittable 4 reloc-right-half)
		(findidnumber  (cadr label)))
	((eq (car label) 'xidloc) 
		(updatebittable 1 reloc-word)
		(updatebittable 3 reloc-xidloc)
		(findidnumber (cadr label)))
	((and (eq (car label) 'high) (numberp (cadr label)))
		 (wshift (cadr label) -11))
	((and (eq (car label) 'low) (numberp (cadr label)))
                        (wshift (wshift (cadr label) 21) -21))
        ((eqcar label 'immediate) (fasladdressvalue (cadr label)))
        ((and (eqcar label 'mkitem)
              (eqcar (caddr label) 'idloc))
                 (mkitem (cadr label) (idinf (cadr (caddr label)))))
        ((and (eqcar label 'mkitem)
              (stringp (caddr label)))
                 (mkitem (cadr label) (fasladdressvalue (caddr label))))
        ((and (eqcar label 'mkitem) (codep (caddr label))) 
              (mkitem (cadr label)  (inf (caddr label))))
	(t    (error 0 (list "Undefined label:" label)))))

(de lapaddressvalue (label)
  (cond ((numberp label) label)
        ((assoc label label-alist*)
                (cdr (assoc label label-alist*)) )
        ((atom label)  (error 0 (list "Undefined label:" label)))
        ((eq (car label) 'plus) (plus (lapaddressvalue (cadr label))
                                (lapaddressvalue (caddr label))))
        ((and (eq (car label) 'high) (numberp (cadr label)))
                        (wshift (cadr label) -11))
        ((and (eq (car label) 'low) (numberp (cadr label)))
                        (wshift (wshift (cadr label) 21) -21))
        ((and (eq (car label) 'high) (eqcar (cadr label) 'immediate))
                        (wshift (lapaddressvalue (cadr label)) -11))
        ((and (eq (car label) 'low) (eqcar (cadr label) 'immediate))
                        (wand (lapaddressvalue (cadr label)) 16#7ff))
	((and (eq (car label) 'high) (pairp (cadr label))
		 (memq (caadr label) '(FLUID $FLUID GLOBAL $GLOBAL)))
		(wshift (lapaddressvalue (cadr label)) -11))
        ((and (eq (car label) 'low) (pairp (cadr label))
                 (memq (caadr label) '(FLUID $FLUID GLOBAL $GLOBAL)))
			(wand (lapaddressvalue (cadr label)) 16#7ff))
	((eq (car label) 'high) (wshift (wplus2 codebase*
                                        (lapaddressvalue (cadr label))) -11))
        ((eq (car label) 'low) (wand (wplus2 codebase*
 				     (lapaddressvalue (cadr label))) 16#7ff))
        ((memq (car label) '(FLUID $FLUID GLOBAL $GLOBAL))
               (wplus2 SYMVAL (wtimes2 4 (id2int (cadr label)))))
        ((memq (car label) '(entry foreignentry))
              (wplus2 SYMFNC (wtimes2 4  (id2int (cadr label)))))
        ((eq (car label) 'idloc)  (id2int (cadr label)))
        ((eq (car label) 'xidloc) (wtimes2 4 
					 (id2int (cadr label)) ))
        ((eqcar label 'immediate) (lapaddressvalue (cadr label)))
        ((and (eqcar label 'mkitem)
              (eqcar (caddr label) 'idloc))
                 (mkitem (cadr label) (idinf (cadr (caddr label)))))
        ((and (eqcar label 'mkitem) (stringp (caddr label)))
              (mkitem (cadr label) (wplus2 codebase*
					 (lapaddressvalue (caddr label)))))
        ((and (eqcar label 'mkitem) (codep (caddr label)))
              (mkitem (cadr label)  (inf (caddr label))))
        (t    (error 0 (list "Undefined label:" label)))))
 

(de jumpaddressvalue (label)
  (cond (*writingfaslfile (fasljumpaddressvalue label))
	(t (lapjumpaddressvalue label))))

(de lapjumpaddressvalue (label)
  (cond ((numberp label) label)
	((assoc label label-alist*)
              (wdifference (cdr (assoc label label-alist*)) currentoffset*))
	((atom label)  (error 0 (list "Undefined label:" label)))
	((eq (car label) 'foreignentry)
	      (getmem (wplus2 SYMFNC (wtimes2 4 (id2int (cadr label))))))
	(t    (error 0 (list "Undefined label:" label)))))

(define-constant reloc-call 2)

(de fasljumpaddressvalue (label)
  (cond ((numberp label) label)
	((assoc label label-alist*)
%       (updatebittable 1 reloc-word)
%       (updatebittable 3 reloc-call)
	       (wdifference (cdr (assoc label label-alist*)) currentoffset*))
	((atom label)  (error 0 (list "Undefined label:" label)))
	((eq (car label) 'internalentry)
	     %(updatebittable 1 reloc-word)
	     %(updatebittable 3 reloc-call) 
	     (setq label (cadr label))  
	     (cond ((assoc label label-alist*)
	              (wdifference (cdr (assoc label label-alist*)) 
				currentoffset*))
		   ((get label 'internalentryoffset)
		      (wdifference (get label 'internalentryoffset)
 				currentoffset*))
		   (t (push `(call ,label ,currentoffset*)
		                forwardinternalreferences*) 0)))
	(t  (error 0 (list "Undefined label:" label)))))

(de forward-internal-reference-p (label)
  (not (or (assoc label label-alist!*)
	   (get label 'internalentryoffset))))

(de code-print (x)
  (for (from i 0 7)
       (do (printf "%x" (wand (wshift  x (wtimes2 4 (wdifference i 7))) 15)))))

%%      store-xxxx is used for saving code into BPS space. 

(de store-word (word)     (putmem (plus2 codebase* currentoffset*) word))

(de store-byte (byte)     (depositbyte byte))

(de store-word-expression (word)  (depositwordexpression word))

(de store-double (word)           (depositfloat `(FLOAT ,word)))

(de store-relocatable-movex (code)   (deposit-relocatable-movex code))

(de putbittable (baseaddress bitoffset value2)
  (prog (m b c)
        (setq b
              (iland (byte baseaddress (setq m (ilsh bitoffset -2)))
                     (ilsh (idifference -1 (itimes2 3 256))
                      (idifference -2
                       (setq c (itimes2 (field bitoffset 30 2) 2))))))
        (putbyte baseaddress m (if (eq value2 0)
                   b
                   (ilor b (ilsh value2 (idifference 6 c)))))))
