%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           SPARC-REGISTERS.SL
% Title:          SPARC PSL register allocation                                
% Author:         Winfried Nuen , ZIB Berlin 
%                 derived from Tsuyoshi Yamamoto's Sparc-lap.sl
% Created:        9 Nov 1988
% Status:         Experimental
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%	Define register names
%%
(deflist '((r0 0)(r1 1)(r2 2)(r3 3)(r4 4)(r5 5)(r6 6)(r7 7)
	   (r8 8)(r9 9)(r10 10)(r11 11)(r12 12)(r13 13)(r14 14)
	   (r15 15)(r16 16)(r17 17)(r18 18)(r19 19)(r20 20)
	   (r21 21)(r22 22)(r23 23)(r24 24)(r25 25)(r26 26)
	   (r27 27)(r28 28)(r29 29)(r30 30)(r31 31))
  '*register-code)

(deflist '((f0 0)(f1 1)(f2 2)(f3 3)(f4 4)(f5 5)(f6 6)(f7 7)
	   (f8 8)(f9 9)(f10 10)(f11 11)(f12 12)(f13 13)(f14 14)
	   (f15 15)(f16 16)(f17 17)(f18 18)(f19 19)(f20 20)
	   (f21 21)(f22 22)(f23 23)(f24 24)(f25 25)(f26 26)
	   (f27 27)(f28 28)(f29 29)(f30 30)(f31 31))
  '*f-register-code)

(deflist '((c0 0)(c1 1)(c2 2)(c3 3)(c4 4)(c5 5)(c6 6)(c7 7)
	   (c8 8)(c9 9)(c10 10)(c11 11)(c12 12)(c13 13)(c14 14)
	   (c15 15)(c16 16)(c17 17)(c18 18)(c19 19)(c20 20)
	   (c21 21)(c22 22)(c23 23)(c24 24)(c25 25)(c26 26)
	   (c27 27)(c28 28)(c29 29)(c30 30)(c31 31))
  '*c-register-code)
%%
%%
%%	Other names of registers.
%%
(deflist
  '((g0 0)(g1 1)(g2 2)(g3 3)(g4 4)(g5 5)(g6 6)(g7 7)
    (o0 8)(o1 9)(o2 10)(o3 11)(o4 12)(o5 13)(o6 14)(o7 15)
    (l0 16)(l1 17)(l2 18)(l3 19)(l4 20)(l5 21)(l6 22)(l7 23)
    (i0 24)(i1 25)(i2 26) (i3 27)(i4 28)(i5 29)(i6 30)(i7 31))
  '*register-code)
%%
%%	Psl friendly names of registers.
%%
(deflist
  '((argone r1)(argtwo r2)(argthree r3)(argfour r4)(argfive r5)
    (argsix 6) (argseven 7) (argeight 8) (argnine 9) (argten 10)
    (argeleven 11) (argtwelve 12) (argthirteen 13) (argfourteen 14)
    (argfivteen 15) (argsixteen 16)
    (NIL r6) (st r14) (sp r14) (unix-st r14) (returnaddr r31)
    (t1 r8) (t2 r9) (t3 r10) (t4 r11) (t5 r12) (t6 r13)) '*other-name)
%%
%%
%%
(de expand-operand (asm-code)
  (cond ((atom asm-code) asm-code)
	(t   (cons (car asm-code)
	       (mapcar (cdr asm-code)
		       (function (lambda(x)(to-standard-name x))))))))

(global '(*REGISTER-NAMES*))

(setq *REGISTER-NAMES*
  '((0 r0)(1 r1)(2 r2)(3 r3)(4 r4)(5 r5)(6 r6)(7 r7)(8 r8)(9 r9)
    (10 r10)(11 r11)(12 r12)(13 r13)(14 r14)(15 r15)(16 r16)
    (17 r17)(18 r18)(19 r19)(20 r20)(21 r21)(22 r22)(23 r23)
    (24 r24)(25 r25)(26 r26)(27 r27)(28 r28)(29 r29)(30 r30)
    (31 r31)))

(de to-standard-name (u)
  (cond ((idp u) (cond ((get u '*other-name) (get u '*other-name))
	                (t u)))
	((pairp u)
	 (cond ((eq (car u) 'REG)
                    (cond ((numberp (cadr u))
                            (cadr (assoc (cadr u) *REGISTER-NAMES*)))
                         ((idp (cadr u)) (to-standard-name (cadr u)))
                         (t (print-error-status)
		            (error 0 (list "Unknown SPARC register name" u)))))
	       ((member (car u) '(MKITEM WARRAY)) u)
	       ((eq (car u) 'DEFERRED)
		(list (to-standard-name (cadr u))))
	       ((member (car u) '(INDEXED DISPLACEMENT))
		(list (to-standard-name (cadr u))
		      (to-standard-name (caddr u))))
	       ((eq (car u) '+) (cadr u))
	       ((eq (car u) '-) u)
	       (t (mapcar u
			(function to-standard-name)))))
	(t u)))

(de sparc-register-p (r)
  (cond ((idp r)
         (cond ((get r '*other-name)
                (sparc-register-p (get r '*other-name)))
               ((get r '*register-code) t)
               ((get r '*f-register-code) t)
               ((get r '*c-register-code) t)
               (t nil)))
        ((eqcar r 'REG) t)
        (t nil)))

% Adapting a useful scheme from Cray-PSL:
%    Some fluid or global variables are heavily used ,some others are almost
%    unused. So it makes sense to store the value cells of those heavy used
%    vars in special locations where we have fast access to them.
%    This is done in  T/B registers on Cray1/X/Y and in local memory on Cray2.
%    On Sparc its useful to store such vars 'around' a fixed pointer (e.g. g7).
%    With that access method defined there are 2**13 bytes = 2048 items of 
%    memory direct addressable via  %g7 + offset.
%    Therefore we have 1 instruction to load / store the values instead of
%    three (sethi, and, ld)
%    'Extragregs' (Reg 6 ..reg 15) are located here too.
%    Also nice for a cache machine.
%
%    Definition below reads as follows:
%    name  number   will be used to change  (*move (fluid name) (reg any)) 
%                                   into    (ld %g7+number (reg any))
%    done by fluid/global anyreg
%
%    Convention:  PSL uses negative offsets, applications e.g. REDUCE use
%    positive offsets > 100. (Reg 6 .. Reg 15 are located in 0 ..39) 
%
%    Garbage collector has to mark from and update this array.

(deflist '((Stacklowerbound  -4)
           (Stackupperbound  -8)
           (Heapupperbound  -12)
           (Heaplowerbound  -16)
           (&symflag&       -20)
           (Arithargs0      -24)
           (Arithargs1      -28)
           (Heaptrapped     -32)
           (T               -36)
           (Betahi*         -40)
           (Betalow*        -44)
           (currentscantable* -48)
           (iobuffer        -52)
           (Nargreg         -56)
           (nextposition    -60)
           (lineposition    -64)
           (writefunction   -68)
           (readfunction    -72)
           (maxline         -76)
           (chtokentype     -80)
           (currentchar     -84)
           (bufferlength    -88)
           (tokch           -92)
           (unreadbuffer    -96)
           (lambda-link-address -100)
           (codeptr*        -104)
           (undefined-function-address -108)
           (catchstackptr     -112)
           (SYMVAL            -116)
           (SYMPRP            -120)
           (SYMGET            -124)
           (SYMNAM            -128)
           (SYMFNC            -132)
           (Hashtable         -136)
           (Bndstkptr         -140)
           (Tokenbuffer       -144)
           (heaplast          -148)
           (heaptrapbound     -152)
           (bndstkupperbound  -156)
           (argumentblock     -160)
           (bndstklowerbound  -164)
           (codeargs          -168)
           (lambindargs*      -172)
           (in*               -176)
           (out*              -180)
           (*echo             -184)
           (PSLDAT            -188)   %  %g7 as PSL variable
         
           (*writingfaslfile  -200)   % some friends from compiler
           (*writingasmfile   -204)
           (*pcmac            -208)
           (*pcode            -212)
           (*pgwd             -216)
           (*plap             -220) 
           (currentoffset*    -224)    
           (*asmpass*         -228)
           (*pwrds            -232)
           (*object*          -236)
           (*outputmode*      -240)
           (channeltable      -244)
           (*second-value*    -248)

           (gcarraylowerbound -252)
           (gcarrayupperbound -256)
           (oldheaplast       -260)
           (oldheaptrapbound  -264) 

           (progjumptable*    -268)
           (progbody*         -272)

           (alglist*          -276)
           (kord*             -280)

           (carry*            -284)
           (oldheaplowerbound -292)
           (oldheapupperbound -296)
           (auxviv0	      -300)
           (auxviv1	      -304)
                                    ) 'VIV-number)

