%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CRAY-CMAC-3.SL - CRAY PSL cmacros , Part 3                            %
%                                                                       %
% Author:      M. L. Griss, S. Lowder, R. Kessler                       %
% (PSL3.2)     Computer Science Dept.                                   %
%              University of Utah                                       %
% Date:        1 June 1982                                              %
%                                                                       %
% Based on M68K-CMAC.SL  1 June by Lowder et al                         %
%              changed nearly all for Cray PSL 3.4                      %
%              by H. Melenk and W. Neun,ZIB Berlin                      %
%                                                                       %
% Copyright (c) 1987, Konrad Zuse Zentrum Berlin, all rights reserved   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions   20 June1988 , W. Neun
%    changed recognition of masks such that
%    bignums are not treated as masks and fixnums only if
%    to temporary register (e.g. arithmetic workhorse uses it)
% Revisions   10 May 1988 , W. Neun
%    changed Bstack over(underflow) calls to be aller jamais retour.
%    that makes code simpler. (one jump less in main line).
%---------------------------------------------------------------------
%   Create and item from the tag stored in the first argument and the
%    information (inf) in the second.
%   FORMAT: InformationAndDestination Tag
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% MkItem may be used when evaluating Wconst expressions.                %
% An item is formed on the CRAY with the first 5 bits as the Tag,       %
%   ----------------------------                                        %
%   | Tag |    Info field      |                                        %
%   ----------------------------                                        %
%  0      5   27               63                                       %
%                                                                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (flag '(foreignlink1 c-foreignlink1) 'internalfunction))
 
(DE MKITEM (TAGPART INFPART) (LOR (LSH TAGPART 59) INFPART))
 
(DefCMacro *MkItem
         ((AnySRegButS7P AnySRegButS7P)
                        (*Move NonTagPart (Reg A6))
                        (DRShift (Reg S7) ArgOne (Reg A6))
                        (Move ArgOne ArgTwo)
                        (DLShift ArgOne (Reg S7) (Reg A6)))
         ((SRegP SRegP)   (*Move NonTagPart (Reg A6))
                        (Move (Reg A7) ArgTwo)
                        (DRShift ArgTwo ArgOne (Reg A6))
                        (Move ArgOne (Reg A7))
                        (DLShift ArgOne ArgTwo (Reg A6))
                        (Move ArgTwo (Reg A7)))
         ((AnySRegButS7P)  (*Move NonTagPart (Reg A6))
                        (DRShift (Reg S7) ArgOne (Reg A6))
                        (*Move ArgTwo ArgOne)
                        (DLShift ArgOne (Reg S7) (Reg A6)))
         ((SRegP)       (*Move NonTagPart (Reg A6))
                        (DRShift (Reg S6) ArgOne (Reg A6))
                        (*Move ArgTwo ArgOne)
                        (DLShift ArgOne (Reg S6) (Reg A6)))
         (              (*Move ArgOne (Reg S6))
                        (*MkItem (Reg S6) ArgTwo)
                        (*Move (Reg S6) ArgOne)))
 
%---------------------------------------------------------------------
%   Move the Source to the Destination Source -> Destination.
%   FORMAT: Source Destination
(DefCMacro *Move
         (Equal)
%%%      ((IndexedP  TRegP)       (*wplus3 (Reg A0)
%%%                                  (pickoffindexedReg ArgOne)
%%%                                  (pickoffindexamt ArgOne))
%%% x-mp ?                        (move (Reg A6) 1 )
%%%                               (Blklod (Reg a6) ArgTwo))
%%%      ((TRegp IndexedP)        (*wplus3 (Reg A0)
%%%                                  (pickoffindexedReg ArgTwo)
%%%                                  (pickoffindexamt ArgTwo))
%%%  x-mp ?                       (move (Reg A6) 1 )
%%%                               (Blksto (Reg a6) ArgOne))
%%%     ((AsmModeP  TRegP)        (move (Reg a0) ArgOne)
%%% x-mp?                         (move (Reg a6) 1)
%%%                               (Blklod (Reg a6) ArgTwo))
%%%     ((TRegP AsmModeP)         (move (Reg a0) ArgTwo )
%%% x-mp ?                        (move (Reg a6) 1)
%%%                               (Blksto (Reg A6) ArgOne))
         ((SRegP TRegP)           (move ArgTwo ArgOne))
         ((TRegP SRegP)           (move ArgTwo ArgOne))
         ((TRegP Anyp)            (move (Reg S6) ArgOne)
 %% caution never change this from s6! otherwise progbind is dead!!
                                  (*Move (Reg s6) ArgTwo))
% move cdr's different from other load operations to gain
% overlapped load (hopefully)
         ((CdrP TRegP)            (*Move ArgOne (Reg S7))
                                  (*Move (Reg S7) ArgTwo))
         ((Anyp TRegp)            (*Move ArgOne (Reg S6))
                                  (*Move (Reg S6) ArgTwo))
         ((ARegP BRegP)           (move ArgTwo ArgOne))
         ((BRegP ARegP)           (move ArgTwo ArgOne))
         ((BRegP Anyp)            (move (Reg A6) ArgOne)
                                  (*Move (Reg A6) ArgTwo))
         ((Anyp BRegp)            (*Move ArgOne (Reg A6))
                                  (*Move (Reg A6) ArgTwo))
         ((RegP   RegP)           (move ArgTwo ArgOne))
         ((Rmaskp SRegp)          (*MoveMask Argone Argtwo))
         ((Rmaskpfixnum TempSRegp)(*MoveMask Argone Argtwo))
         ((SmallInumP ASRegP)     (move ArgTwo ArgOne))
         ((SmallInumP IndexedP)   (move (Reg S6) ArgOne)
                                  (store (Reg S6) ArgTwo))
         ((QuotedIdp SRegp)       (*Movequotedid ArgOne ArgTwo)
                                  (*move 4 (Reg A6))
                                  (Drshift ArgTwo ArgTwo (Reg A6)))
         ((QuotedIdp Anyp)        (*Move ArgOne (Reg S6))
 %% caution never change this from s6! otherwise progbind is dead!!
                                  (*Move (Reg S6) ArgTwo ))
    % Don't try too hard to understand the above code!!
    % It is curios: an single id won't get loaded
    % but is constructed (without memory-reference) .
         ((FixP ASRegP)           (load ArgTwo (lit (fullword ArgOne))))
         ((FixP IndexedP)         (load (Reg S6) (lit (fullword ArgOne)))
                                  (store (Reg S6) ArgTwo))
         ((FloatP SRegP)          (load ArgTwo (quote ArgOne)))
         ((FloatP IndexedP)       (load (Reg S6) (quote ArgOne))
                                  (store (Reg S6) ArgTwo))
         ((ImmediateP AnyP)       (*Loc ArgTwo ArgOne))
         ((IndexedP ASRegP)       (load ArgTwo ArgOne))
         ((ASRegP IndexedP)       (store ArgOne ArgTwo))
         ((IndexedP IndexedP)     (load (Reg s6) ArgOne)
                                  (store (Reg s6) ArgTwo))
         ((QuoteP ASRegP)         (load ArgTwo ArgOne))
         ((QuoteP AnyP)           (*MOVE ArgOne (Reg S7))
                                  (*MOVE (Reg S7) ArgTwo))
         ((AsmModeP ASRegP)       (load ArgTwo ArgOne))
         ((ASRegP AsmModeP)       (store ArgOne ArgTwo))
         ((AsmModeP AsmModeP)     (load (Reg s6) ArgOne)
 %% caution never change this from s6! otherwise lambind is dead!!
                                  (store (Reg s6) ArgTwo))
         ((AsmModeP AnyP)         (load (Reg s6) ArgOne)
                                  (*MOVE (Reg s6) ArgTwo))
         ((AnyP AsmModeP)         (*MOVE ArgOne (Reg s6))
                                  (store (Reg s6) ArgTwo))
         ((ParcelP ASRegP)        (move ArgTwo ArgOne))
      % parcelp in Crinstr
         (          (*Cerror "Weird Move" ArgOne ArgTwo)))
 
% Support for move
 
(Defcmacro *movequotedid)
 
(de *movequotedid (ArgOne ArgTwo)
        `((move ,ArgTwo (xidloc ,(cadr ArgOne) 15))))
 
(Defcmacro *movemask)
 
(de *movemask (ArgOne ArgTwo)
   (prog (x)
      (setq x (Rmaskp argone))
      (when (not x) (setq x (Rmaskpfixnum Argone)))
      (return `((rmask ,ArgTwo ,x)))))
 
(de Rmaskp(x)
     (when (eqcar x 'quote) (setq x (cadr x)))
     (if (and (intp x) (not (eq x 0)))
                            (powerof2p (wplus2 x 1))
         NIL))
 
(de Rmaskpfixnum (x)
     (when (eqcar x 'quote) (setq x (cadr x)))
     (if (and (fixnp x) (not (eq x 0)))
                     (powerof2p (wplus2 (int2sys x) 1))
         NIL))
 
%(de Rmaskp(x)
%     (when (eqcar x 'quote) (setq x (cadr x)))
%     (if (and (fixp x) (not (bigp x)) (not (eq x 0)))
%                  (powerof2p (wplus2 (int2sys x) 1 )) nil))
 
(de powerof2p (x)
       (if (eq 1 (wpopcnt x)) (wdifference 63 (wldzero x)) nil))
 
% Definition of MOVESGN
 
(DefCMacro *Movesgn)
 
(de *Movesgn (ArgOne ArgTwo) 
   (setq argtwo (resolveoperand '(reg t1) argtwo))
   (cond ((and (aregp argone)
               (sregp argtwo))    `((Movesgn ,ArgTwo ,ArgOne)))
         ((aregp argone)          `((movesgn (reg s6) ,argone)
                                    (*move (reg s6) ,argtwo)))
         (t                       `(*move ,argone ,argtwo))))

 
%---------------------------------------------------------------------
%   Pop the top element of the stack into the Destination
%   FORMAT: Destination
 
(DefCMacro *Pop
         (    (*CERROR " POP should not be called !! " 'Pop argone)))
 
%---------------------------------------------------------------------
%   Place into the Destination a Field of the Operand Starting at BitPosition
%    (from left starting at zero), of size Length
% ^^WRONG^^ FORMAT: Destination Operand BitPosition Length
%^^^^^^^^^^ Above format is incorrect.  Documentation needs to be changed.
%^^^^^^^^^^  Actual format is:
%   FORMAT: Operand Destination BitPosition Length
%?????????? Assume for now that BitPosition and Length are Constants
%?????????? Problem^^ - For PUTFIELD to work we need three Registers.
%??????????  Therefore, I don't see how we can make this work, without
%??????????  producing a grundle of macros, testing to see which Registers
%??????????  are in use, and saving the contents of another Register to
%??????????  make room.  What this needs is one spare Register that it can
%??????????  use.  It is ok to save the contents of the Register before hand,
%??????????  but somehow, we need to pick another Register to use.  Any
%??????????  Ideas??  I guess that I can write a LISP routine that given
%??????????  the two arguments, it selects and returns a Register to use.
%??????????  Functionality of the routine is that if Reg S6 or S7 is free
%??????????  then a move is not necessary, otherwise save a Register into
%??????????  T7 and then restore.  We could use the fact that *Move will
%??????????  NOT generate code if the same, then we can call a routine to
%??????????  check if they are s6 and s7, if not then return t16, which
%??????????  will not generate anything.
%********** Here, working on other options.
% The following two cmacros must be defined, because we need to call a
% function within their definition (The call to useRegister below).  That
% way, we can expand the function call and just generate the code.
 
(DefCMacro *RMask)
 
(de *RMask (ArgOne ArgTwo)
      `((rmask ,(resolveoperand '(Reg t1) ArgOne) ,Argtwo)))
 
(DefCMacro *Merge
           (            (Merge ArgOne ArgTwo ArgThree)))
 
(de *Merge (ArgOne ArgTwo ArgThree ArgFour)
  (Expand4OperandCMacro ArgOne ArgTwo ArgThree ArgFour '*Merge))
 
(fluid '(getarg3*))
 
(de getarg3 (x) (setq getarg3* x))
 
(de getarg4 (x) (equal (wplus2 getarg3* x) 64))
 
(DefCMacro *PutField
      ((SRegP SRegP getarg3 getarg4)  % optimize for inf-field case
                     (*Move (ChooseRegister ArgTwo ArgOne) (Reg T7))
                     (*RMask (UseRegister ArgTwo ArgOne)  ArgFour)
                     (*Merge ArgTwo ArgOne
                                          (UseRegister ArgTwo ArgOne))
                     (*Move (Reg T7)  (ChooseRegister ArgTwo ArgOne)))
      ((SRegP SRegP) (*Move (ChooseRegister ArgTwo ArgOne) (Reg T7))
                     (*Move (Difference 64
                                 (Plus2 ArgThree ArgFour)) (Reg A6))
                     (DRShift ArgTwo ArgTwo (Reg A6))
                     (*RMask (UseRegister ArgTwo ArgOne)  ArgFour)
                     (*Merge ArgTwo ArgOne
                             (UseRegister ArgTwo ArgOne))
                     (DLShift ArgTwo ArgTwo (Reg A6))
                     (*Move (Reg T7) (ChooseRegister ArgTwo ArgOne)))
      ((AnyP AnySRegButS7P)(*Move ArgOne (Reg S7))
                        (*PutField (Reg S7) ArgTwo ArgThree ArgFour))
      ((AnyP SRegP)      (*Move ArgOne (Reg S6))
                         (*PutField (Reg S6) ArgTwo ArgThree ArgFour))
      ((AnySRegButS7P AnyP)(*Move ArgTwo (Reg S7))
                         (*PutField ArgOne (Reg S7) ArgThree ArgFour)
                         (*Move (Reg S7) ArgTwo))
      ((SRegP AnyP)      (*Move ArgTwo (Reg S6))
                         (*PutField ArgOne (Reg S6) ArgThree ArgFour)
                         (*Move (Reg S6) ArgTwo))
      (                  (*Move ArgTwo (Reg S6))
                         (*Move ArgOne (Reg S7))
                         (*PutField (Reg S7) (Reg S6) ArgThree ArgFour)
                         (*Move (Reg S6) ArgTwo)))
 
%---------------------------------------------------------------------
%   Push the argument onto the stack.
%   FORMAT: Argument
(DefCMacro *Push
         (  (*Cerror "Push should not be called!!" 'PUSH ArgOne)))
 
%---------------------------------------------------------------------
%   Extract the *Field (same arguments) and sign extend it.
%   FORMAT: Destination Operand BitPosition Length
%?????????? Optimize for the possible redundant move, if ArgOne is not
%?????????? in an SRegister.
(DefCMacro *SignedField
         (          (*Field ArgOne ArgTwo ArgThree ArgFour)
                    (*SignExtend ArgOne (Difference 64 ArgFour))))
 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 AND Arg2
%   FORMAT: Destination^-Operand1 Operand2
(DefCMacro *WAND
         (          (*WAND3 ArgOne ArgOne ArgTwo)))
 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 - Arg2
%   FORMAT: Destination^-Operand1 Operand2
(DefCMacro *WDIFFERENCE
         (           (*WDIFFERENCE3 ArgOne ArgOne ArgTwo)))
 
%---------------------------------------------------------------------
%   Leading zero
%
(De *wldzero(Argone argtwo)
  (Expand2OperandCMacro ArgOne Argtwo '*Wldzero))
 
(DefCmacro *wldzero
       ((AnyP Sregp) (zeropop (Reg A6) ArgTwo)
                     (*move (Reg a6) argone))
       (             (*move argtwo (Reg S6))
                     (*wldzero argone (Reg S6))))
 
%---------------------------------------------------------------------
%   Population count
%
(De *wpopcnt (Argone Argtwo)
  (Expand2OperandCMacro ArgOne Argtwo '*Wpopcnt))
 
(DefCmacro *wpopcnt
       ((Anyp Sregp) (onepop (Reg A6) argTwo)
                     (*move (Reg a6) argone))
       (             (*move argTwo (Reg S6))
                     (*wpopcnt argone (Reg S6))))
 
%---------------------------------------------------------------------
%   Population count parity
%
(De *wpopcntp (Argone ArgTwo)
  (Expand2OperandCMacro ArgOne ArgTwo '*Wpopcntp))
 
(DefCmacro *wpopcntp
       ((Anyp Sregp) (paritypop (Reg A6) argTwo)
                     (*move (Reg a6) argone))
       (             (*move argTwo (Reg S6))
                     (*wpopcntp argone (Reg S6))))
 
%---------------------------------------------------------------------
%   Arg1 <- - Arg2 (two's complement)
%   FORMAT: Destination Operand1
%?????????? This is inconsistent with the other W operators.  They are
%??????????  all two address, where this is three address (i.e. specifies
%??????????  destination explicitly.
(DefCMacro *WMinus
        ((ABRegP ABRegP)  (*WDIFFERENCE3 ArgOne (Reg A0) ArgTwo))
        (                 (*WDIFFERENCE3 ArgOne (Reg S0) ArgTwo)))
 
%---------------------------------------------------------------------
%   Arg1 <- NOT Arg2 (one's complement)
%   FORMAT: Destination Operand1
%?????????? This is inconsistent with the other W operators.  They are
%??????????  all two address, where this is three address (i.e. specifies
%??????????  destination explicitly.
(DefCMacro *WNot
         ((SRegP SRegP)      (NOT ArgOne ArgTwo))
         ((SRegP)            (*MOVE ArgTwo (Reg S7))
                             (*WNot ArgOne (Reg S7)))
         ((AnyP  AnySRegButS7P) (*WNot (Reg S7) ArgTwo)
                                (*MOVE (Reg S7) ArgOne))
         ((AnyP  SRegP) (*WNot (Reg S6) ArgTwo)
                        (*MOVE (Reg S6) ArgOne))
         (              (*MOVE ArgTwo (Reg S7))
                        (*WNot (Reg S7) (Reg S7))
                        (*MOVE (Reg S7) ArgOne)))
 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 OR Arg2
%   FORMAT: Destination^-Operand1 Operand2
(DefCMacro *WOR
         (              (*WOR3 ArgOne ArgOne ArgTwo)))
 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 + Arg2
%   FORMAT: Destination^-Operand1 Operand2
(DefCMacro *WPLUS2
         (              (*WPLUS3 ArgOne ArgOne ArgTwo)))
 
%---------------------------------------------------------------------
%   Logical shift with zero fill in either end.  A positive shift index
%    is a left shift, while a negative one is to the right.
%   FORMAT: OperandToShift BitsToShift
(DefCMacro *WShift                  %Logical shift. +index=left.
         ((AnyP WZeroP))
         ((SRegP PosLE64P)   (LSHIFT ArgOne ArgOne ArgTwo))
         ((SRegP NegLE64P)   (RSHIFT ArgOne ArgOne (Minus ArgTwo)))
         ((SRegP ARegP)      (*JUMPWLESSP TEMPLabel ArgTwo 0)
                             (LSHIFT ArgOne ArgOne ArgTwo)
                             (*JUMP TEMPLabel2)
                             (*LBL (Label TEMPLabel))
                             (*WMINUS ArgTwo ArgTwo)
                             (RSHIFT ArgOne ArgOne ArgTwo)
                             (*LBL (Label TEMPLabel2)))
         ((SRegP)            (*MOVE ArgTwo (Reg A6))
                             (*WShift ArgOne (Reg A6)))
         (                   (*MOVE ArgOne (Reg s6))
                             (*WShift (Reg s6) ArgTwo)
                             (*MOVE (Reg s6) ArgOne)))
 
%---------------------------------------------------------------------
%   Logical shift with zero fill in either end.
%   This is an unconditional leftshift !!!
%   FORMAT: OperandToShift BitsToShift
(DefCMacro *WLShift
         ((AnyP WZeroP))
         ((SRegP PosLE64P)   (LSHIFT ArgOne ArgOne ArgTwo))
         ((SRegP ARegP)
                             (LSHIFT ArgOne ArgOne ArgTwo))
         ((SRegP)            (*MOVE ArgTwo (Reg A6))
                             (*WLShift ArgOne (Reg A6)))
         (                   (*MOVE ArgOne (Reg s6))
                             (*WLShift (Reg s6) ArgTwo)
                             (*MOVE (Reg s6) ArgOne)))
 
 
(de *wlshift (ArgOne ArgTwo)
         (expand2operandcmacro ArgOne ArgTwo '*wlshift))
 
%---------------------------------------------------------------------
%   Logical shift with zero fill in either end.
%   This is an unconditional rightshift !!!
%   FORMAT: OperandToShift BitsToShift
(DefCMacro *WRShift
         ((AnyP WZeroP))
         ((SRegP PosLE64P)   (RSHIFT ArgOne ArgOne ArgTwo))
         ((SRegP ARegP)
                             (RSHIFT ArgOne ArgOne ArgTwo))
         ((SRegP)            (*MOVE ArgTwo (Reg A6))
                             (*WRShift ArgOne (Reg A6)))
         (                   (*MOVE ArgOne (Reg s6))
                             (*WRShift (Reg s6) ArgTwo)
                             (*MOVE (Reg s6) ArgOne)))
 
(de *wrshift (ArgOne ArgTwo)
         (expand2operandcmacro ArgOne ArgTwo  '*wrshift))
 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 * Arg2    Multiplication for addresses
%   FORMAT: Destination^-Operand1 Operand2 (24 bits ^^^^)
(DefCMacro *wTimes2
         ((AnyP OneP)        )
         ((AnyP WZeroP) (*Move 0 ArgOne))
         ((aregp aregp) (MUL ArgOne ArgOne ArgTwo))
         (              (*Move ArgOne (Reg A6))
                        (*Move ArgTwo (Reg A7))
                        (MUL (Reg A6) (Reg A6) (Reg A7))
                        (MOVESGN (Reg S6) (Reg A6))
                        (*Move (Reg S6) ArgOne)))
%---------------------------------------------------------------------
%   Arg1 <- Arg1 / Arg2
%   FORMAT: Destination^-Operand1 Operand2
(de *WQuotient (dividend divisor)
  (let ((divisor
       % Strip off the QUOTE on quoted numbers (hope this is safe, be a
       % bit careful about what kind of numbers we treat like this).
    (cond
     ((and (eqcar divisor 'QUOTE) (FixP (cadr divisor))) (cadr divisor))
     ((and (eqcar divisor 'Wconst)(FixP (cadr divisor))) (cadr divisor))
      (T    divisor))))
    (cond
      ((and (AnyP dividend) (OneP divisor))
       % No-op if divide by one
       NIL)
      ((and (AnySRegButs7P dividend) (PowerOf2P divisor))
       % Note that PowerOf2P returns the base 2 logarithm, if not NIL.
       `((*Move ,dividend (Reg S7))
         (*WRshift (Reg S7) ,(Difference 64 (Powerof2p divisor)))
         (*WPlus2 ,dividend (Reg S7))
         (*ASHIFT ,dividend ,(minus (PowerOf2P divisor)))))
      ((and (SRegP dividend) (PowerOf2P divisor))
       % Note that PowerOf2P returns the base 2 logarithm, if not NIL.
       `((*Move ,dividend (Reg S6))
         (*WRshift (Reg S6) ,(Difference 64 (Powerof2p divisor)))
         (*WPlus2 ,dividend (Reg S6))
         (*ASHIFT ,dividend ,(minus (PowerOf2P divisor)))))
      ((and (AnyP dividend) (PowerOf2P divisor))
       % Note that PowerOf2P returns the base 2 logarithm, if not NIL.
       `((*Move ,dividend (Reg S6))
         (*Move (Reg s6) (Reg S7))
         (*WRshift (Reg S6) ,(Difference 64 (Powerof2p divisor)))
         (*WPlus2 (reg S7) (Reg S6))
         (*ASHIFT (Reg s7) ,(minus (PowerOf2P divisor)))
         (*move (reg S7) ,dividend)))
      (T 
       % This is the hairy part^^
       (cond
         ((and (ARegP dividend) (ARegP divisor))
        % Follow the CAL manual, do integer division by forming the
          % floating point reciprocal approximation and multiplying by that.
        `( % Move the A Registers into S Registers, floating them in the
           % process.
           (MOVEFL (Reg S6) ,dividend)
           (MOVEFL (Reg S7) ,divisor)
           % Normalize them by adding 0.
           (FADD (Reg S6) (Reg S0) (Reg S6))
           (FADD (Reg S7) (Reg S0) (Reg S7))
           % Get the reciprocal of the divisor, to about 30 bits.
           (HRECIP (Reg S7) (Reg S7))
           % Multiply dividend by the reciprocal to get quotient in S6.
           (HMUL (Reg S6) (Reg S6) (Reg S7))
           % S7 gets magic number to fix (unfloat) things.
           (SCONST06 (Reg S7))
           (FSUB (Reg S6) (Reg S7) (Reg S6))
           (SUB (Reg S6) (Reg S7) (Reg S6))
           % Finally, replace the original dividend with the quotient.
           (*MOVE (Reg S6) ,dividend)))
       ((ARegP dividend)
        % If just the dividend is in an A Register...
        `( % Move the divisor into A7--a scratch A Register.
          (*MOVE ,divisor (Reg A7))
           % Now, get the quotient, don't need to move the result.
          (*Wquotient ,dividend (Reg A7))))
       ((ARegP divisor)
        % If just the divisor is in an A Register...
        `((*MOVE ,dividend (Reg A6))
          (*Wquotient (Reg A6) ,divisor)
           % Replace the dividend with the quotient.
          (*MOVESGN (Reg A6) ,dividend)))
       (T
        % Neither argument is in an A Register.
        `((*MOVE ,dividend (Reg A6))
          (*MOVE ,divisor  (Reg A7))
          (*Wquotient (Reg A6) (Reg A7))
           % Replace the dividend with the quotient.
          (*MOVESGN (Reg A6) ,dividend))))))))
 
(DefCMacro *WQuotient)
 
(commentoutcode   % older version
(DefCMacro *cons
      ((AnySRegButS7P SRegP) (*Move (Reg Heaplast) (Reg a6))
                     (*Move 2 (Reg a7))
                     (Add (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaptrapbound) (Reg a5))
                     (store ArgOne 0 (Reg a6))
                     (store ArgTwo 1 (Reg a6))
                     (Sub (Reg a0) (Reg a5) (Reg a7))
                     (*Move (Reg a7) (Reg Heaplast))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) (Reg S7))
                     (*wshift ArgOne 59)
                     (*wor ArgOne (Reg S7))
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
      % cons-reclaim is defined in compacting-gc
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel)))
      ((SRegP SRegP) (*Move (Reg Heaplast) (Reg a6))
                     (*Move 2 (Reg a7))
                     (*Move ArgTwo (Reg T1))
                     (Add (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaptrapbound) (Reg a5))
                     (store ArgOne 0 (Reg a6))
                     (store ArgTwo 1 (Reg a6))
                     (Sub (Reg a0) (Reg a5) (Reg a7))
                     (*Move (Reg a7) (Reg Heaplast))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) ArgTwo)
                     (*wshift ArgOne 59)
                     (*wor ArgOne ArgTwo)
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel))
                     (*move (Reg T1) ArgTwo))
      ((AnyP AnySRegButS7P)(*Move ArgOne (Reg S7))
                           (*Cons (Reg S7) ArgTwo )
                           (*Move (Reg S7) ArgOne))
      ((AnyP SRegP)        (*Move ArgOne (Reg S6))
                           (*cons (Reg S6) ArgTwo )
                           (*Move (Reg S6) ArgOne))
      ((AnySRegButS7P AnyP)(*Move ArgTwo (Reg S7))
                           (*Cons ArgOne (Reg S7)))
      ((SRegP AnyP)        (*Move ArgTwo (Reg S6))
                           (*cons ArgOne (Reg S6)))
      (                    (*Move ArgTwo (Reg S6))
                           (*Move ArgOne (Reg S7))
                           (*cons (Reg S7) (Reg S6))
                           (*Move (Reg S7) ArgOne))
)) %end commentoutcode
 
(de *cons (ArgOne ArgTwo)
    (expand2operandcmacro ArgOne ArgTwo '*cons))
 
(commentoutcode   % older version
(DefCMacro *xcons
      ((AnySRegButS7P SRegP) (*Move (Reg Heaplast) (Reg a6))
                     (*Move 2 (Reg a7))
                     (Add (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaptrapbound) (Reg a5))
                     (store ArgOne 1 (Reg a6))
                     (store ArgTwo 0 (Reg a6))
                     (Sub (Reg a0) (Reg a5) (Reg a7))
                     (*Move (Reg a7) (Reg Heaplast))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) (Reg S7))
                     (*wshift ArgOne 59)
                     (*wor ArgOne (Reg S7))
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
      % cons-reclaim is defined in compacting-gc
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel)))
      ((SRegP SRegP) (*Move (Reg Heaplast) (Reg a6))
                     (*Move 2 (Reg a7))
                     (*Move ArgTwo (Reg T1))
                     (Add (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaptrapbound) (Reg a5))
                     (store ArgOne 1 (Reg a6))
                     (store ArgTwo 0 (Reg a6))
                     (Sub (Reg a0) (Reg a5) (Reg a7))
                     (*Move (Reg a7) (Reg Heaplast))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) ArgTwo)
                     (*wshift ArgOne 59)
                     (*wor ArgOne ArgTwo)
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel))
                     (*move (Reg T1) ArgTwo))
      ((AnyP AnySRegButS7P)(*Move ArgOne (Reg S7))
                           (*xcons (Reg S7) ArgTwo )
                           (*Move (Reg S7) ArgOne))
      ((AnyP SRegP)        (*Move ArgOne (Reg S6))
                           (*xcons (Reg S6) ArgTwo )
                           (*Move (Reg S6) ArgOne))
      ((AnySRegButS7P AnyP)(*Move ArgTwo (Reg S7))
                           (*xCons ArgOne (Reg S7)))
      ((SRegP AnyP)        (*Move ArgTwo (Reg S6))
                           (*xcons ArgOne (Reg S6)))
      (                    (*Move ArgTwo (Reg S6))
                           (*Move ArgOne (Reg S7))
                           (*xcons (Reg S7) (Reg S6))
                           (*Move (Reg S7) ArgOne))
)
) % end commentoutcode
 
(de *xcons (ArgOne ArgTwo)
    (expand2operandcmacro ArgOne ArgTwo '*xcons))
 
(DefCMacro *cons
      ((AnySRegButS7P SRegP) (*Move (Reg Heaptrapbound) (Reg a6))
                     (*Move 2 (Reg a7))
                     (SUB (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaplast) (Reg a5))
                     (store ArgOne 10 (Reg a6))
                     (store ArgTwo 11 (Reg a6))
                     (Sub (Reg a0) (Reg a7) (Reg a5))
                     (*Move (Reg a7) (Reg Heaptrapbound))
                     (*move 10 (reg a7))
                     (*wplus2 (reg a6) (reg a7))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) (Reg S7))
                     (*wshift ArgOne 59)
                     (*wor ArgOne (Reg S7))
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
      % cons-reclaim is defined in compacting-gc
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel)))
      ((SRegP SRegP) (*Move (Reg Heaptrapbound) (Reg a6))
                     (*Move 2 (Reg a7))
                     (*Move ArgTwo (Reg T1))
                     (SUB (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaplast) (Reg a5))
                     (store ArgOne 10 (Reg a6))
                     (store ArgTwo 11 (Reg a6))
                     (Sub (Reg a0) (Reg a7) (Reg a5))
                     (*Move (Reg a7) (Reg Heaptrapbound))
                     (*move 10 (reg a7))
                     (*wplus2 (reg a6) (reg a7))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) ArgTwo)
                     (*wshift ArgOne 59)
                     (*wor ArgOne ArgTwo)
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel))
                     (*move (Reg T1) ArgTwo))
      ((AnyP AnySRegButS7P)(*Move ArgOne (Reg S7))
                           (*Cons (Reg S7) ArgTwo )
                           (*Move (Reg S7) ArgOne))
      ((AnyP SRegP)        (*Move ArgOne (Reg S6))
                           (*cons (Reg S6) ArgTwo )
                           (*Move (Reg S6) ArgOne))
      ((AnySRegButS7P AnyP)(*Move ArgTwo (Reg S7))
                           (*Cons ArgOne (Reg S7)))
      ((SRegP AnyP)        (*Move ArgTwo (Reg S6))
                           (*cons ArgOne (Reg S6)))
      (                    (*Move ArgTwo (Reg S6))
                           (*Move ArgOne (Reg S7))
                           (*cons (Reg S7) (Reg S6))
                           (*Move (Reg S7) ArgOne))
)
 
(DefCMacro *xcons
      ((AnySRegButS7P SRegP) (*Move (Reg Heaptrapbound) (Reg a6))
                     (*Move 2 (Reg a7))
                     (SUB (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaplast) (Reg a5))
                     (store ArgOne 11 (Reg a6))
                     (store ArgTwo 10 (Reg a6))
                     (Sub (Reg a0) (Reg a7) (Reg a5))
                     (*Move (Reg a7) (Reg Heaptrapbound))
                     (*move 10 (reg a7))
                     (*wplus2 (reg a6) (reg a7))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) (Reg S7))
                     (*wshift ArgOne 59)
                     (*wor ArgOne (Reg S7))
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
      % cons-reclaim is defined in compacting-gc
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel)))
      ((SRegP SRegP) (*Move (Reg Heaptrapbound) (Reg a6))
                     (*Move 2 (Reg a7))
                     (*Move ArgTwo (Reg T1))
                     (SUB (Reg a7) (Reg a6) (Reg a7))
                     (*Move (Reg Heaplast) (Reg a5))
                     (store ArgOne 11 (Reg a6))
                     (store ArgTwo 10 (Reg a6))
                     (Sub (Reg a0) (Reg a7) (Reg a5))
                     (*Move (Reg a7) (Reg Heaptrapbound))
                     (*move 10 (reg a7))
                     (*wplus2 (reg a6) (reg a7))
                     (*Move 9 ArgOne)
                     (*Move (Reg a6) ArgTwo)
                     (*wshift ArgOne 59)
                     (*wor ArgOne ArgTwo)
                     (JAP (Label tempLabel))
                     (*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
                     (*move (Reg A7) (Reg B00))
                     (*lbl (Label tempLabel))
                     (*move (Reg T1) ArgTwo))
      ((AnyP AnySRegButS7P)(*Move ArgOne (Reg S7))
                           (*xcons (Reg S7) ArgTwo )
                           (*Move (Reg S7) ArgOne))
      ((AnyP SRegP)        (*Move ArgOne (Reg S6))
                           (*xcons (Reg S6) ArgTwo )
                           (*Move (Reg S6) ArgOne))
      ((AnySRegButS7P AnyP)(*Move ArgTwo (Reg S7))
                           (*xcons ArgOne (Reg S7)))
      ((SRegP AnyP)        (*Move ArgTwo (Reg S6))
                           (*xcons ArgOne (Reg S6)))
      (                    (*Move ArgTwo (Reg S6))
                           (*Move ArgOne (Reg S7))
                           (*xcons (Reg S7) (Reg S6))
                           (*Move (Reg S7) ArgOne))
)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% fast mkfltn; fast allocation of a floating point number and
%              putting the float value; clearing register for gc
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(DefCMacro *fastmkfltn
   ((AnySRegButS7P) (*Move (Reg Heaplast) (Reg a6))
                    (*Move 2 (Reg a7))
                    (Add (Reg a7) (Reg a6) (Reg a7))
                    (*Move (Reg Heaptrapbound) (Reg a5))
                    (store ArgOne 1 (Reg a6))
                    (*move 25 ArgOne)             %hwords-tag
                    (*wshift ArgOne 59)
                    (store ArgOne 0 (Reg a6))
                    (Sub (Reg a0) (Reg a5) (Reg a7))
                    (*Move (Reg a7) (Reg Heaplast))
                    (*Move 3 ArgOne)              %float-tag
                    (*Move (Reg a6) (Reg S7))
                    (*wshift ArgOne 59)
                    (*wor ArgOne (Reg S7))
                    (JAP (Label tempLabel))
                    (*move (Reg B00) (Reg a7))
                    (*call cons-reclaim)
              % cons-reclaim is defined in compacting-gc
                    (*move (Reg A7) (Reg B00))
                    (*lbl (Label tempLabel)))
 
   (                (*move ArgOne (Reg S6))
                    (*fastmkfltn (Reg S6))
                    (*Move (Reg S6) ArgOne))
)
 
(de *fastmkfltn (ArgOne)
    (expand1operandcmacro ArgOne '*fastmkfltn))
 
 
% Cmacro to call reclaim without a visible link
%
(de *reclaim ()
                   '((*move (Reg B00) (Reg a7))
                     (*call cons-reclaim)
      % cons-reclaim is defined in compacting-gc
                     (*move (Reg A7) (Reg B00))))
 
(defcmacro *reclaim)
 
%---------------------------------------------------------------------
%   Arg1 <- Arg1 Xor Arg2
%   FORMAT: Destination^-Operand1 Operand2
(DefCMacro *WXOr
         (                  (*WXOR3 ArgOne ArgOne ArgTwo)))
 
%========================================================================
(fluid '(Labelgen*))            % a-list of tags and Labels
 
% (Labelgen tag) and (Labelref tag) can be used as either ANYReg or CMACRO.
% (Labelgen tag) creates and returns a unique Label, (Labelref tag) returns
% the same one.  Useful for 'OpenCode lists.
(commentoutcode
(de myanyRegLabelgen (Reg name)
  ((lambda (lb al)
     (cond ((null al) (setq Labelgen* (cons (cons name lb) Labelgen*)))
           (t (rplacd al lb)))
         (list 'parcel lb))
   (gensym)
   (assoc name Labelgen*)))
 
(defanyReg myLabelgen myanyRegLabelgen)
 
(de myanyRegLabelref (Reg name) (cdr (assoc name Labelgen*)))
 
(defanyReg myLabelref myanyregLabelref)
)
 
(De *LamBind (Regs Fluids)
 
 % be careful, code is somewhat tricky
 
 (prog (n lng list cregs cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Regs (rest Regs))   % Remove "Registers" from the front
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Regs) 2)) % two words per BndStk entry
 
   %looking for free register to preload values of fluid s
 
      (setq freeregs (mapcan '((reg 1)(reg 2)(reg 3)(reg 4)(reg 5))
                       (function (lambda (x)
                                    (cond ((member x regs) nil)
                                           (t (cons x nil))
      )              ) )         )  )
      (setq freeregs (nconc (list '(reg S6) '(reg s7)) freeregs))
      (setq cfluids fluids) % copy of fluids
 
preload  (setq initload
           (progn (setq cadrcfluids
                    (nconc cadrcfluids (cons (car freeregs) nil)))
            (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))
 
       (cond ((and freeregs cfluids) (go preload)))
 
       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist
 
      (setq list `((*move (Reg BndStkPtr) (Reg A1))
                   (*move ,lng  (Reg A7))
                   (*wplus2 (Reg A7) (Reg A1))
                   (*move (Reg BndstkUpperBound) (Reg A5))
%                  (*jumpwlessp (Label ,lab) (Reg A7) (Reg A5))
%                  (*call Bstackoverflow)
                   (*wdifference3 (reg a0) (reg a7)(reg a5))
                   (jap (plus2 (parcel symfnc)
                               (times2 (idloc Bstackoverflow) 4)))
%                  (*Lbl (Label ,lab))
                   (*move (Reg A7) (Reg BndstkPtr))  )) %start of code
 
      (setq list (append initload list))
 
  loop
      (setq cregs (car Regs))
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
               (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 2))
      (Setq list (append list
                 `((*move ,(car freeregs) (indexed (Reg A1) ,n))
                   (*move (quote ,Cadrcfluids)
                               (indexed (reg A1) ,(wplus2 n -1)))
                   (*move ,cregs ,cfluids)
      )          ))
      (setq fluids (cdr Fluids))
      (setq freeregs (cdr freeregs))
      (cond ((setq regs (cdr Regs)) (go loop)))
      (return list)
)    )
(defcmacro *lambind)
 
(De *ProgBind (Fluids)
 
 % be careful, code is somewhat tricky
 
 (prog (n lng list cfluids cadrcfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
 
      (setq freeregs '((reg S6)(reg s7)(reg 1)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
 
preload  (setq initload
           (progn (setq cadrcfluids
                    (nconc cadrcfluids (cons (car freeregs) nil)))
            (nconc initload `((*move ,(car cfluids) ,(car freeregs))))
       )   )
       (setq freeregs (cdr freeregs))
       (setq cfluids (cdr cfluids))
 
       (cond ((and freeregs cfluids) (go preload)))
 
       (setq freeregs (nconc cadrcfluids cfluids)) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and not preloaded fluids if those exist
 
 
      (setq list `((*move (Reg BndStkPtr) (Reg A1))
                   (*move ,lng  (Reg A7))
                   (*wplus2 (Reg A7) (Reg A1))
                   (*move (Reg BndstkUpperBound) (Reg A5))
%                  (*jumpwlessp (Label ,lab) (Reg A7) (Reg A5))
%                  (*call Bstackoverflow )
                   (*wdifference3 (reg a0) (reg a7)(reg a5))
                   (jap (plus2 (parcel symfnc)
                               (times2 (idloc Bstackoverflow) 4)))
%                  (*Lbl (Label ,lab))
                   (*move (Reg  A7) (Reg BndstkPtr))  )) %start of code
 
     (setq list (append initload list))
 
 loop
      (setq cfluids (car Fluids))
      (setq cadrcfluids (cadr cfluids))
      (when (or (eq cadrcfluids 't) (eq cadrcfluids 'nil))
               (stderror "T and NIL cannot be rebound"))
      (setq n (wplus2 n 2))
      (Setq list (append list
                 `((*move ,(car freeregs) (indexed (Reg A1) ,n))
                   (*move (quote ,Cadrcfluids)
                               (indexed (reg A1) ,(wplus2 n -1)))
                   (*move (quote nil) ,cfluids)
      )          ))
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *progbind)
 
(De *FreeRstr (Fluids)
 
 (prog (n lng list cfluids listfluids lab initload freeregs)
      (setq lab (gensym))
      (setq n 0)
      (setq Fluids (rest fluids)) % Remove NONLOCALVARS
      (setq lng (wtimes2 (length Fluids) 2)) % two words per BndStk entry
      (setq freeregs '((reg S6)(reg s7)(reg 2)(reg 3)(reg 4)(reg 5)))
      (setq cfluids fluids) % copy of fluids
      (setq n (wdifference 2 lng))
      (setq initload (list '(*move (reg Bndstkptr) (reg a6))))
 
preload  (setq initload
           (progn (setq listfluids
                   (if freeregs
                    (nconc listfluids (cons (car freeregs) nil))
                    (nconc listfluids (cons nil nil))) )
            (nconc initload
             (if freeregs
                `((*move (indexed (reg a6) ,n) ,(car freeregs))) nil)
       )   ))
       (setq n (wplus2 n 2))
       (when freeregs (setq freeregs (cdr freeregs)))
       (setq cfluids (cdr cfluids))
 
       (cond (cfluids (go preload)))
 
       (setq freeregs listfluids ) %end of preloadcode
 
       % freeregs contains the list of preloaded regs
       % and nil if not enough regs available
 
      (setq list `((*move ,lng (Reg A7))
                   (*wdifference3 (Reg A5) (Reg A6) (Reg A7))
                   (*move (Reg BndstkLowerBound) (Reg A7))
%                  (*jumpwgeq (Label ,lab) (Reg A5) (Reg A7))
%                  (*call Bstackunderflow)
                   (*wdifference3 (reg a0) (reg a5)(reg a7))
                   (jam (plus2 (parcel symfnc)
                               (times2 (idloc Bstackunderflow) 4)))
%                  (*Lbl (Label ,lab))
                   (*move (Reg A5) (Reg BndstkPtr))  )) %start of code
 
     (setq list (append initload list))
     (setq n 0)
 
 loop
      (setq cfluids (car Fluids))
      (setq n (wplus2 n 2))
 
  % insert reloaded register or memory reference
 
      (setq list (append list
               (if (car freeregs) `((*move ,(car freeregs) ,cfluids ))
                            `((*move (indexed (Reg A5) ,n) ,cfluids )))
      )          )
      (setq freeregs (cdr freeregs))
      (cond ((setq Fluids (cdr Fluids)) (go loop)))
      (return list)
)    )
(defcmacro *freerstr)
 
(setq *unsafebinder t)   % has to save Registers across calls
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                  MISC. DEFINITIONS                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(DEFCMACRO *Jumpon)
(De *Jumpon (Register Lowerbound Upperbound Labellist)
    (PROG (X LL LL2)
       (setq ll  (gensym))
       (setq ll2 (gensym))
       (SETQ X
        (if (and (weq lowerbound 0) (weq upperbound 31) *syslisp)
                  % jumpon on tags (most probably)
          `((*move ,register (reg s7))
            (lshift (reg s7) (reg s7) 1)
            (ploc (reg a7) (Label ,ll2))
            (*move (reg s7) (reg a6))
            (*wplus2 (reg a6) (reg a7))
            (*move (reg A6) (reg b07))
            (j (reg b07))
            (*lbl (Label ,ll2)))
          `((Move (Reg S7) (Wconst ,UPPERBOUND))
            (Sub  (Reg S0) ,Register (Reg S7))
            (Move (Reg S7) (Wconst ,LOWERBOUND))
            (Sub  (Reg S7) ,Register (Reg S7))
            (Jsz  ,(LASTCAR LabelLIST))
            (Jsp  (Label ,ll))
            (Move (Reg S0) (Reg S7))
            (Jsm  (Label ,LL))
            (Lshift (Reg S7) (Reg S7) (Wconst 1))
            (Move (Reg A6) (Reg S7))
            (Ploc (Reg A7) (Label ,ll2))
            (Add (Reg A7) (Reg A7) (Reg A6))
            (Move (Reg B07) (Reg A7))
            (J (Reg B07))
            (*Lbl (Label ,LL2))
       )) )
      Loop  (Setq x (nconc X `((J ,(car Labellist))) ))
            (setq Labellist (cdr Labellist))
            (cond (Labellist (go loop)))
 
            (setq x (nconc x `((*lbl (Label ,ll))) ))
            (return x)
)     )
 
% Foreign Call cross assembly only
 
(Fluid '(*cft-mode))
 
(put 'TRegsave 'symbol 'TRegsave)
(put 'BRegsave 'symbol 'BRegsave)
 
(de *foreignlink (fn ft noa)
  (append
     '((*Move (Reg A3) ($Fluid A3save))
       (*Move (Reg St) ($Fluid A4save))
       (Move  (Reg A0) BRegsave)
       (*Move  64 (Reg A1))
       (Blksto (Reg A1) (Reg B00))
       (Move  (Reg A0) TRegsave)
       (Blksto (Reg A1) (Reg T00))
       (Move  (Reg A0) (Plus BRegsave 64))
       (Blklod (Reg A1) (Reg B00))
       (*Move (Reg A1) ($fluid *cft-mode))
      )
   (append (*foreignlink1 fn ft noa)
     '((Move  (Reg A0) (Plus BRegsave 64))
       (*Move  64 (Reg A1))
       (Blksto (Reg A1) (Reg B00))
       (Move  (Reg A0) TRegsave)
       (Blklod (Reg A1) (Reg T00))
       (Move  (Reg A0) BRegsave)
       (Blklod (Reg A1) (Reg B00))
       (*Move (QUOTE NIL) ($fluid *cft-mode))
       (*Move ($fluid &&ForeignArg1) (Reg 1))
       (*Move ($Fluid A3save) (Reg A3))
       (*Move ($Fluid A4save) (Reg St))
      )
)  ))
 
(de *foreignlink1 (fn ft noa)
 
  % stupid code , i regret
 
  (cond ((eq noa 0)  `((Move (Reg A6) (ForeignEntry Arglst0))
                       (r (Foreignentry ,fn))) )
        ((eq noa 1)  `((*Move (Reg 1) ($fluid &&ForeignArg1))
                       (Move (Reg A6) (ForeignEntry Arglst1))
                       (r (Foreignentry ,fn))) )
        ((eq noa 2)  `((*Move (Reg 1) ($fluid &&ForeignArg1))
                       (*Move (Reg 2) ($fluid &&ForeignArg2))
                       (Move (Reg A6) (ForeignEntry Arglst2))
                       (r (Foreignentry ,fn))) )
        ((eq noa 3)  `((*Move (Reg 1) ($fluid &&ForeignArg1))
                       (*Move (Reg 2) ($fluid &&ForeignArg2))
                       (*Move (Reg 3) ($fluid &&ForeignArg3))
                       (Move (Reg A6) (ForeignEntry Arglst3))
                       (r (Foreignentry ,fn))) )
        ((eq noa 4)  `((*Move (Reg 1) ($fluid &&ForeignArg1))
                       (*Move (Reg 2) ($fluid &&ForeignArg2))
                       (*Move (Reg 3) ($fluid &&ForeignArg3))
                       (*Move (Reg 4) ($fluid &&ForeignArg4))
                       (Move (Reg A6) (ForeignEntry Arglst4))
                       (r (Foreignentry ,fn))) )
        (t    (*Cerror "FOREIGNFUNCTION With >4 Parameters" fn noa))
)   )
 
(DefCmacro *Foreignlink)
 
(de *C-foreignlink (fn ft noa)
  (append
     `((*Move (Reg A3) ($Fluid A3save))
       (*Move (Reg St) ($Fluid A4save))
       (Move  (Reg A0) BRegsave)
       (*Move  64 (Reg A1))
       (Blksto (Reg A1) (Reg B00))
       (Move  (Reg A0) TRegsave)
       (Blksto (Reg A1) (Reg T00))
       (Move  (Reg A0) (Plus BRegsave 64))
       (Blklod (Reg A1) (Reg B00))
       (*Move (Reg A1) ($fluid *cft-mode))
       (*Move ($fluid cosargs) (reg a6))
       (*move (quote ,noa) (indexed (reg a6) 0))
      )
   (append (*C-foreignlink1 fn ft noa)
     '((Move  (Reg A0) (Plus BRegsave 64))
       (*Move  64 (Reg A1))
       (Blksto (Reg A1) (Reg B00))
       (Move  (Reg A0) TRegsave)
       (Blklod (Reg A1) (Reg T00))
       (Move  (Reg A0) BRegsave)
       (Blklod (Reg A1) (Reg B00))
       (*Move (QUOTE NIL) ($fluid *cft-mode))
       (*Move ($Fluid A3save) (Reg A3))
       (*Move ($Fluid A4save) (Reg St))
      )
)  ))
 
(de *C-foreignlink1 (fn ft noa)
 
  % stupid code , i regret
 
  (cond ((eq noa 0)  `((r (foreignentry ,fn))) )
        ((eq noa 1)  `((*move (reg 1) (indexed (reg a6) 1))
                       (r (Foreignentry ,fn))) )
        ((eq noa 2)  `((*Move (reg 1) (indexed (reg a6) 1))
                       (*Move (Reg 2) (indexed (reg a6) 2))
                       (r (Foreignentry ,fn))) )
        ((eq noa 3)  `((*Move (Reg 1) (indexed (reg a6) 1))
                       (*Move (Reg 2) (indexed (reg a6) 2))
                       (*Move (Reg 3) (indexed (reg a6) 3))
                       (r (Foreignentry ,fn))) )
        ((eq noa 4)  `((*Move (Reg 1) (indexed (reg a6) 1))
                       (*Move (Reg 2) (indexed (reg a6) 2))
                       (*Move (Reg 3) (indexed (reg a6) 3))
                       (*Move (Reg 4) (indexed (reg a6) 4))
                       (r (Foreignentry ,fn))) )
        (t    (*Cerror "C-FOREIGNFUNCTION With >4 Parameters" fn noa))
)   )
 
(DefCmacro *C-Foreignlink)
