%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           i860-registers.sl
% Title:          Intel i860 PSL register allocation
% Author:         Winfried Neun , ZIB Berlin 
% Created:        9 May 1992
% Mode:           Lisp
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%	Define register names
%%
(deflist '((r0 0)(r1 1)(r2 2)(r3 3)(r4 4)(r5 5)(r6 6)(r7 7)
	   (r8 8)(r9 9)(r10 10)(r11 11)(r12 12)(r13 13)(r14 14)
	   (r15 15)(r16 16)(r17 17)(r18 18)(r19 19)(r20 20)
	   (r21 21)(r22 22)(r23 23)(r24 24)(r25 25)(r26 26)
	   (r27 27)(r28 28)(r29 29)(r30 30)(r31 31))
  '*register-code)

(deflist '((f0 0)(f1 1)(f2 2)(f3 3)(f4 4)(f5 5)(f6 6)(f7 7)
           (f8 8)(f9 9)(f10 10)(f11 11)(f12 12)(f13 13)(f14 14)
           (f15 15)(f16 16)(f17 17)(f18 18)(f19 19)(f20 20)
           (f21 21)(f22 22)(f23 23)(f24 24)(f25 25)(f26 26)
           (f27 27)(f28 28)(f29 29)(f30 30)(f31 31))
  '*f-register-code)
%%
%%	Psl friendly names of registers.
%%
(deflist
  '((argone r16)(argtwo r17)(argthree r18)(argfour r19)(argfive r20)
    (NIL r9) (st r2) (sp r2) (unix-st r2) (returnaddr r1)
    (null r0) (zero r0) 
    (bndstkptr r4) (bndstklowerbound r5) (bndstkupperbound r6)
    (ff1 r21) (ff2 r22) (ff3 r23) (ff4 r24)
    (symfnccenter r7) (symvalcenter r8)
    (t1 r25) (t2 r26) (t3 r27) (t4 r28) (t5 r29) (t6 r30)) '*other-name)

(put 'nil '*other-name 'r9) %doesnt work in deflist ??
(put 't   '*other-name 'r10)

(deflist
  '(    (heaplast r11) (heaptrapbound r12) (t r10)
        (bndstkptr r4) (bndstklowerbound r5) (bndstkupperbound r6)
        ) 'Registername)

(de expand-operand (asm-code)
  (cond ((atom asm-code) asm-code)
        (t   (cons (car asm-code)
               (mapcar (cdr asm-code)
                       (function (lambda(x)(to-standard-name x))))))))

(global '(*REGISTER-NAMES*))
 
(setq *REGISTER-NAMES*  '((0 r0)(1 r16)(2 r17)(3 r18)(4 r19)(5 r20)))
 
(de to-standard-name (u)
  (cond ((idp u) (cond ((get u '*other-name) (get u '*other-name))
                        (t u)))
        ((pairp u)
         (cond ((eq (car u) 'REG)
                    (cond ((numberp (cadr u))
                            (cadr (assoc (cadr u) *REGISTER-NAMES*)))
                         ((idp (cadr u)) (to-standard-name (cadr u)))
                         (t (error 0 (list "Unknown i860 register name" u)))))
               ((member (car u) '(MKITEM WARRAY $fluid $global)) u)
               ((eq (car u) 'DEFERRED)
			(list (car u) (to-standard-name (cadr u))))
               ((member (car u) '(INDEXED DISPLACEMENT))
                (list (car u) (to-standard-name (cadr u))
                      (to-standard-name (caddr u))))
               ((eq (car u) '+) (cadr u))
               ((eq (car u) '-) u)
               (t (mapcar u (function to-standard-name)))))
        (t u)))
 
(de i860-register-p (r)
  (cond ((idp r)
         (cond ((get r 'Registername) T)
               ((get r '*other-name) (i860-register-p (get r '*other-name)))
               ((get r '*register-code) t)
               ((get r '*f-register-code) t)
               ((get r '*c-register-code) t)
               (t nil)))
        ((eqcar r 'REG) t)
        (t nil)))
 
(de get-register-code (reg)
      (cond ((pairp reg) (get-register-code (cadr reg)))
            ((and (get reg '*other-name)
                        (get-register-code (get reg '*other-name))))
            ((eq reg 0) 0)
            ((and (numberp reg) (lessp reg 6)) (plus2 reg 15))
            ((get reg '*register-code) (get reg '*register-code))
	    ((get reg '*f-register-code) (get reg '*f-register-code))
            (t (error 0 (list "Wrong register argument" reg))))
)
              
(de regp (r) (eqcar r 'reg))

(de fregp (r) (regp r))
