%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:COPIERS.SL 
% Title:        Functions for copying various data types 
% Author:       Eric Benson 
% Created:      27 August 1981 
% Modified:     19 Oct 1984 0617-PDT (Mark R. Swanson)
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 29-Feb-96 Winfried Neun (for the sake of this special leapyear day).
%  Invented copyfltn in Totalcopy such that alignment of floats is not lost
% 20-jun-89 Winfried Neun, ZIB Berlin,  Added a fast (lapped) copier for Mips
% in favour of copying-gc
% 4-Jul-85 Edited by Connie O'Dell to install fix for uninitialized vectors
%  in totalCopy proposed by Will Galway, to prevent unpleasant consequences
%  of garbage collecting while part of copy structure is uninitialized.
% 19 Oct 1984 0616-PDT (Mark R. Swanson)
%  Modified length calculation in COPYSTRINGTOFROM to take count word out of 
%   consideration
% 18-Jul-84 08:08:34 (Brian Beach)
%  Minor cleanup of TOTALCOPY
% 01-Dec-83 14:26:58 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime (flag '(fast-copier) 'internalfunction))

(lap '((*ENTRY FAST-COPIER EXPR 4)
          (*alloc 17)
          (*move (REG 4) (DISPLACEMENT (REG ST) 32))
          (*move (REG 3) (DISPLACEMENT (REG ST) 8))
          (*wand (REG 1) (reg infmask))
          (*move (REG 1) (DISPLACEMENT (REG ST) 24))
          (*wand (REG 2) (reg infmask))
          (*move (REG 2) (DISPLACEMENT (REG ST) 4))
          (*MOVE 16 (REG 2))
          (*move (REG 3) (REG 1))
          (DIV (REG 3) (REG 2))
          (mfhi (reg 3))
          (*move (REG 3) ($FLUID *SECOND-VALUE*))
          (mflo (reg 1))
          (*move (REG 1) (reg t7))
          (*move ($FLUID *SECOND-VALUE*) (reg 1))
          (*move 1 (reg 3))
          (*wdifference (reg 1) (reg 3))
          (*move (reg 1) (DISPLACEMENT (REG ST) 16))
          (*MOVE (displacement (REG ST) 24) (REG T5))
          (*wshift (reg t5) 2)
          (*MOVE (DISPLACEMENT (REG ST) 4) (REG T6))
          (*wshift (reg t6) 2)
G0008 
          (slt (reg 5)  (reg t7) (REG 3))
	  (BNE (reg 5) (reg 0) (label g0007))
          (*move (displacement (REG T5) 0) (reg 5))
	  (*move (DISPLACEMENT (REG T5) 4) (reg t1))
           (*move (reg 5) (displacement (REG T6) 0))
           (*move (reg T1) (displacement (REG T6) 4))
          (*move (displacement (REG T5) 8) (reg 5))
          (*move (DISPLACEMENT (REG T5) 12) (reg t1))
           (*move (reg 5) (displacement (REG T6) 8))
           (*move (reg T1) (displacement (REG T6) 12))
          (*move (displacement (REG T5) 16) (reg 5))
          (*move (DISPLACEMENT (REG T5) 20) (reg t1))
           (*move (reg 5) (displacement (REG T6) 16))
           (*move (reg T1) (displacement (REG T6) 20))
          (*move (displacement (REG T5) 24) (reg 5))
          (*move (DISPLACEMENT (REG T5) 28) (reg t1))
           (*move (reg 5) (displacement (REG T6) 24))
           (*move (reg T1) (displacement (REG T6) 28))
          (*move (displacement (REG T5) 32) (reg 5))
          (*move (DISPLACEMENT (REG T5) 36) (reg t1))
           (*move (reg 5) (displacement (REG T6) 32))
           (*move (reg T1) (displacement (REG T6) 36))
          (*move (displacement (REG T5) 40) (reg 5))
          (*move (DISPLACEMENT (REG T5) 44) (reg t1))
           (*move (reg 5) (displacement (REG T6) 40))
           (*move (reg T1) (displacement (REG T6) 44))
          (*move (displacement (REG T5) 48) (reg 5))
          (*move (DISPLACEMENT (REG T5) 52) (reg t1))
           (*move (reg 5) (displacement (REG T6) 48))
           (*move (reg T1) (displacement (REG T6) 52))
          (*move (displacement (REG T5) 56) (reg 5))
          (*move (DISPLACEMENT (REG T5) 60) (reg t1))
           (*move (reg 5) (displacement (REG T6) 56))
           (*move (reg T1) (displacement (REG T6) 60))
          (*MOVE 64 (REG 2))
          (*wplus2 (reg t5) (reg 2))
          (*wplus2 (reg t6) (reg 2))
          (j G0008)
          (*wplus2 (REG 3) 1)
G0007          (*MOVE 0 (REG 2))
          (*move  (DISPLACEMENT (REG ST) 16) (reg 3)) 
G0013          (*move (REG 2) (REG 1))
          (slt (reg 5)  (REG 3) (REG 1))
          (bne  (reg 5) (reg 0) (label g0012))
          (*move (DISPLACEMENT (REG T5) 0) (reg 5))
          (*move (reg 5) (DISPLACEMENT (REG T6) 0))
          (*wplus2 (reg t5) 4)
          (*wplus2 (reg t6) 4)
          (j G0013)
          (*wplus2 (REG 2) 1)
G0012     (*move (DISPLACEMENT (REG ST) 32) (reg 1))
          (*exit 17)))
  
(de copystringtofrom (new old)
  % Copy all chars in Old to New
  (prog (slen stripnew stripold)
	(setq stripnew (strinf new))
	(setq stripold (strinf old))
	(setq slen (strlen stripold))
	(when (wlessp (strlen stripnew) slen)
	  (setq slen (strlen stripnew)))
        (setq slen (- (strpack slen) 1))
%       (for (from i 0 slen 1)
%              (do (setf (vecitm stripnew i) (vecitm stripold i))))
%       (return new)))
	(return (fast-copier (+ stripold 1) (+ stripnew 1) (+ slen 1) new))))  

(de copystring (s)
  % copy to new heap string
  (prog (s1)
	(setq s1 (gtstr (strlen (strinf s))))
	(copystringtofrom s1 (strinf s))
	(return (mkstr s1))))

(de copybytes (s)
  % copy to new heap string
  (prog (s1)
	(setq s1 (gtstr (strlen (strinf s))))
	(copystringtofrom s1 (strinf s))
	(return (mkbytes s1))))

(de copywarray (new old uplim)
  % copy UpLim + 1 words
% (progn (for (from i 0 uplim 1) (do (setf (wgetv new i) (wgetv old i))))
%         new))
  (fast-copier old new (+ uplim 1) new))  

(de copyvectortofrom (new old)
  % Move elements, don't recurse
  (prog (slen stripnew stripold)
	(setq stripnew (vecinf new))
	(setq stripold (vecinf old))
	(setq slen (veclen stripold))
%       (for % assumes VecLen New has been set
%
%             (from i 0 slen 1)
%             (do (setf (vecitm stripnew i) (vecitm stripold i))))
%       (return new)))
  (return (fast-copier (+ stripold 1) (+ stripnew 1) (+ slen 1) new))))  

(de copyvector (s)
  % Copy to new vector in heap
  (prog (s1)
	(setq s1 (gtvect (veclen (vecinf s))))
	(copyvectortofrom s1 (vecinf s))
	(return (mkvec s1))))

(de copyevectortofrom (new old)
  % Move elements, don't recurse
  (prog (slen stripnew stripold)
	(setq stripnew (evecinf new))
	(setq stripold (evecinf old))
	(setq slen (eveclen stripold))
%       (for % assumes EVecLen New has been set
%
%             (from i 0 slen 1)
%             (do (setf (evecitm stripnew i) (evecitm stripold i))))
%       (return new)))
  (return (fast-copier (+ stripold 1) (+ stripnew 1) (+ slen 1) new))))  

(de copyevector (s)
  % Copy to new EVector in heap
  (prog (s1)
	(setq s1 (gtevect (eveclen (evecinf s))))
	(copyevectortofrom s1 (evecinf s))
	(return (mkevect s1))))

(de copywrdstofrom (new old)
  % Like CopyWArray in heap
  (prog (slen stripnew stripold)
	(setq stripnew (wrdinf new))
	(setq stripold (wrdinf old))
	(setq slen (wrdlen stripold))
%       (for % assumes WrdLen New has been set
%
%             (from i 0 slen 1)
%             (do (setf (wrditm stripnew i) (wrditm stripold i))))
%       (return new)))
  (return (fast-copier (+ stripold 1) (+ stripnew 1) (+ slen 1) new))))  

(de copyhalfwordstofrom (new old)
  % Like CopyWArray in heap
  (prog (slen stripnew stripold)
	(setq stripnew (halfwordinf new))
	(setq stripold (halfwordinf old))
	(setq slen (halfwordlen stripold))
	(for % assumes WrdLen New has been set
 
	      (from i 0 slen 1)
	      (do (setf (halfworditm stripnew i) (halfworditm stripold i))))
	(return new)))

(de copywrds (s)
  % Allocate new WRDS array in heap
  (prog (s1)
	(setq s1 (gtwrds (wrdlen (wrdinf s))))
	(copywrdstofrom s1 (wrdinf s))
	(return (mkwrds s1))))

(de copyhalfwords (s)
  % Allocate new HALFWORDS array in heap
  (prog (s1)
	(setq s1 (gthalfwords (halfwordlen (halfwordinf s))))
	(copyhalfwordstofrom s1 (halfwordinf s))
	(return (mkhalfwords s1))))

% CopyPairToFrom is RplacW, found in EASY-NON-SL.RED
% CopyPair is: car S . cdr S;
% Usual Lisp definition of Copy only copies pairs, is found in EASY-NON-SL.RED

(de totalcopy (s)
  % Unique copy of entire structure blows up on circular structures
  (case (tag s)
    ((pair-tag)      (cons (totalcopy (car s)) (totalcopy (cdr s))))
    ((string-tag)    (copystring s))
    ((vector-tag)    (let* ((strips (vecinf s))
	                    (len (veclen strips))
	                    (ptr (mkvect len)))
	               (for (from i 0 len 1)
	                    (do
	                     (setf (vecitm (vecinf ptr) i)
	                       (totalcopy (vecitm (vecinf s) i)))))
	               ptr))
    ((words-tag)     (copywrds s))
    ((bytes-tag)     (copybytes s))
    ((halfwords-tag) (copyhalfwords s))
    ((fixnum-tag)    (mkfixn (inf (copywrds s))))

    ((floatnum-tag)  (copyfltn s))
    (nil             s)
    ))

(de copyfltn (s)
  % Allocate new Floatnum in heap
  % May have special alignment needs
  (prog (s1)
        (setq s1 (gtfltn (wrdlen (wrdinf s))))
        (copywrdstofrom s1 (wrdinf s))
        (return (mkfltn s1))))


(off fast-integers)
