%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PNK:COMP-SUPPORT.SL
% Title:          Run-time support for assorted things
% Author:         Eric Benson
% Created:        20 September 1981
% Modified:       29-Aug-84 11:01:48 (Brian Beach)
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1987, Konrad Zuse Zentrum Berlin, for modifications
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 19 Sep 88 W. Neun   ZIB Berlin
%            modified list2-list15 for new heap scheme (upper heap)
% 19 Jun 86 H. Melenk ZIB Berlin
%            modified list7-list15 using vector hardware
% 5 Jun 86 W. Neun ZIB Berlin
%            added list6-list15 due to new *cons cmacro
% 21-Apr 86 W. Neun ZIB Berlin,added Idapply0 - idapply5 as smacros
%            changed list1-list5 due to new *cons cmacro
% 01-Dec-83 14:23:53 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%
%   vectorized versions of functions list7 - list15
%   warning: do not change sequence of operations
%   (loss of execution speed!)
%
%   common to all these functions: parameters 1 to 5 are
%   located in the registers
%
%   parameters 6 - 15 lie in the memory at location ArgumentBlock
%   as a technical vector
%
 
 
(compiletime (load numeric-ops vector-instructions))
 
(compiletime (on fast-integers))
 
(compiletime
  (progn 
                   % number or argument for a LISTnn function
   (define-constant number 15)
 
                             % macros for vector-load instructions
   (dm  list-load-cdr-vector (u)
    (cons 'progn (list-load-cdr-form 0
                                 (- (compiler-constant 'number) 1))))
 
   (de list-load-cdr-form (n m)
    (if (eq n m)
        nil
        (cons
           `(VsetElement (vreg 2) ,n ,(* (+ n 1) 2))
            (list-load-cdr-form (+ n 1) m) )))
))                                                  % compiletime
 
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% list15
 
(compiletime (setq number 15))
 
(de list15 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST14 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 14))
 
(de list14 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST13 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 13))
 
(de list13 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST12 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 12))
 
(de list12 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST11 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 11))
 
(de list11 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST10 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 10))
 
(de list10 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST9  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 9))
 
(de list9 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST8  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 8))
 
(de list8 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
 
% LIST7   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (setq number 7))
 
(de list7 (p1 p2 p3 p4 p5    % par 6 - 15 in argumentblock
    )%
   (prog()
      (when ( >=  ( + heaplast (* number 2)) heaptrapbound)
            (!%!%reclaim))
 
                        % read argumentBlock
     (VsetVL (- number 5))
     (Vload (Vreg 1) ArgumentBlock 1)
 
                        % move pars 1 - 5 to Vreg
     (VsetElement (vreg 0) 0 p1)
     (VsetElement (vreg 0) 1 p2)
     (VsetElement (vreg 0) 2 p3)
     (VsetElement (vreg 0) 3 p4)
     (VsetElement (vreg 0) 4 p5)
                        % heap request
     (setq p1 (mkpair (setq p2 (- (+ 12 heaptrapbound) (* number 2)))))
     (setq heaptrapbound (- heaptrapbound (* number 2)))
 
 
     (VsetVL (- number 5))
     (Vstore (Vreg 1) (+ p2 10) 2)  % store second part of cars
 
                        % calculate cdr pointers
     (VsetElement (Vreg 3) (- number 1)  nil)    % last nil
     (list-load-cdr-vector)
     (VsetVL (- number 1))
     (VplusS  (Vreg 3) (Vreg 2) p1)  % add tagged baseaddress
 
     (VsetVL 5)
     (Vstore (Vreg 0) p2 2)          % store first part of cars
 
     (VsetVL number)
     (Vstore (Vreg 3) (+ p2 1) 2)    % store cdrs
 
     (return p1)
))
% NCONS, XCONS are defined in the Kernel
 
(commentoutcode
(de list15 (u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11 u12 u13 u14 u15)
  % 15-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 (cons u10
  (cons u11 (cons u12 (cons u13 (cons u14 (cons u15 nil
  )))))))))))))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list14 (u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11 u12 u13 u14)
  % 14-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 (cons u10
  (cons u11 (cons u12 (cons u13 (cons u14 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list13 (u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11 u12 u13 )
  % 13-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 (cons u10
  (cons u11 (cons u12 (cons u13 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list12 (u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11 u12 )
  % 12-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 (cons u10
  (cons u11 (cons u12 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list11 (u1 u2 u3 u4 u5 u6 u7 u8 u9 u10 u11)
  % 11-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 (cons u10
  (cons u11 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list10 (u1 u2 u3 u4 u5 u6 u7 u8 u9 u10)
  % 10-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 (cons u10 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list9 (u1 u2 u3 u4 u5 u6 u7 u8 u9)
  % 9-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8 (cons u9 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list8 (u1 u2 u3 u4 u5 u6 u7 u8)
  % 8-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 (cons u8  nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(commentoutcode
(de list7 (u1 u2 u3 u4 u5 u6 u7)
  % 7-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 (cons u7 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(de list6 (u1 u2 u3 u4 u5 u6)
  % 6-argument EXPR for LIST
  (cons u1 (cons u2 (cons u3 (cons u4 (cons u5
  (cons u6 nil
  )))))))))))))))))))))))))))))))))))))))))))
 
(de list5 (u v w x y)
  % 5-argument EXPR for LIST
  (cons u (cons v (cons w (cons x (cons y nil))))))
 
(de list4 (u v w x)
  % 4-argument EXPR for LIST
  (cons u (cons v (cons w (cons x  nil)))))
 
(de list3 (u v w)
  % 3-argument EXPR for LIST
  (cons u (cons v (cons w  nil))))
 
(de list2 (u v)
  % 2-argument EXPR for LIST
  (cons u (cons v nil)))
 
(ds idapply0 (u) (idapply u))
 
(ds idapply1 (p1 u) (idapply u (list p1)))
 
(ds idapply2 (p1 p2 u) (idapply u (list p1 p2 )))
 
(ds idapply3 (p1 p2 p3 u) (idapply u (list p1 p2 p3 )))
 
(ds idapply4 (p1 p2 p3 p4 u) (idapply u (list p1 p2 p3 p4)))
 
(ds idapply5 (p1 p2 p3 p4 p5 u) (idapply u (list p1 p2 p3 p4 p5)))
 
 
%
% Support for WARRAYS:
%
 
(on fast-integers)
 
(de clear-warray (address size)
  (for (from index 0 (wplus2 size -1))
       (do (wputv address index 0))
       ))
 
(de initialize-warray (address initial-values)
  (for (in value initial-values)
       (from index 0)
       (do (wputv address index value))
       ))
 
(de clear-wstring (address char-count word-count)
  (clear-warray address word-count)
  (wputv address 0 (wplus2 char-count -1))
  )
 
(de initialize-wstring (address initial-string word-count)
  (let ((inf (strinf initial-string)))
    (for (from i 0 (wplus2 word-count -1))
       (do (wputv address i (wgetv inf i)))
       )))
 
(off fast-integers)
 
