%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:            System-io
% Description:     CRAY-UNICOS specific IO routines for PSL
% Author:          Modified by W.Neun, ZIB Berlin from HP/UX Version by
%                  Russ Fish, Robert R. Kessler, Martin L. Griss
%                  Also for Cray-2.
% Created:         23-Apr-88
% Mode:            Lisp
% Status:          Experimental (Do Not Distribute)
%
% Copyright (c) 1988, Konrad Zuse-Zentrum Berlin for Cray version
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions:
% 17-March-1989 Winfried Neun
% put errorset around opne for /dev/tty since it is not present in NQS batch
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Defintely NOT a final version, BAUSTELLE!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Each individual system must have the following routines defined.
% SysClearIo, SysOpenRead, SysOpenWrite, SysReadRec, SysWriteRec, SysClose,
% SysMaxBuffer
%
%      The following definitions are used in the routines:
%       FileDescriptor - A machine dependent word that references a file once
%                        opened.
%       FileName - A Lisp string of the file name.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime (progn (load io-decls)
                     (flag '(cosstring) 'internalfunction)))
 
(fluid '(cos-stderr* cos-tty* stdout* stdin* promptout* errout* cosnull
))
(de cosstring (s)
  % Strip tags for passing a string.
     (if (stringp s)  (strbase (strinf s))
                      (stderror (bldmsg "Not a string: %p%n" s))))
 
(de unixstring(s) (cosstring s))

%------------------------------------------------------------------------------
% Any system dependent clearing that has to be done.
%------------------------------------------------------------------------------
 
   (de syscleario ()
  % Establish channels #'s for initial system.
     (COS-clrio)

     (setf cosnull      -1)
     (setf stdin*       0)
     (setf stdout*      1)
     (setf promptout*   stdout*)
     (setf errout*      stdout*)
     (setf cos-stderr*  5)
     (setf cos-tty*     6)

  % Hand Build channeltables.  cos-tty* is the controlling tty even when
  % stdin is redirected.
  % The numbers 59, 58 and 57 are historical derived from COS version.
  % should be the true unix pointers.

  (clearonechannel stdin*      59  'input  'terminalinputhandler)
  (clearonechannel stdout*     58  'output 'independentwritechar)
  (clearonechannel cos-stderr* 57  'output 'independentwritechar)
  (setq cosnull (errorset '(sysopenread  0 "/dev/tty") NIL NIL) )
               % prevent open-error when nqs batch jobs
  (clearonechannel cos-tty*  (if (eq cosnull 90) 59 cosnull)
                                   'input  'terminalinputhandler)
  (setf cosnull      -1))
 
%------------------------------------------------------------------------------
% Open FileName for input and return a file descriptor used in later
% references to the file.
%------------------------------------------------------------------------------
 
(de sysopenread (channel filename)
 
    (prog (filed filenam)
        (setq filenam (if (idp filename)
                            (symnam (idinf filename)) filename))
        (setq filed (cos-openr filed (cosstring filenam)))
        (if (wgreaterp filed 0) (return filed)
            (conterror 90  % errno is biased with -10000
                (bldmsg "Open-Error ,errno = %w" (wplus2 filed 10000))
                (sysopenread channel filename))
            (return nil)
)   )   )
 
%------------------------------------------------------------------------------
% Open FileName for output and return a file descriptor used in later
% references to the file.
%------------------------------------------------------------------------------
 
(de sysopenwrite (channel filename)
 
   (prog (filed filenam)
        (setq filenam (if (idp filename)
                            (symnam (idinf filename)) filename))
        (setq filed (cos-openw filed (cosstring filenam)))
        (if (wgreaterp filed 0) (return filed)
            (conterror 91  % errno is biased with -10000
                (bldmsg "Open-Error ,errno = %w" (wplus2 filed 10000))
                (sysopenwrite channel filename))
            (return nil)
)   )   )
 
%------------------------------------------------------------------------------
% Read from the FileDescriptor, putting a line record into the StringBuffer
% terminated by an EOL or EOF character. Return the length of the string read
% as the indexing bound (one less than the number of characters read.)
%------------------------------------------------------------------------------
 
(de sysreadrec (filedescriptor stringbuffer)
 
  (prog (lng)     % Check for reading past the end of file.
    (setq lng (cos-readr (sysmaxbuffer filedescriptor)
                         (cosstring stringbuffer) filedescriptor))
    (return
      (if (wneq lng cosnull)  (wplus2 lng -1)
       (progn
        (setf (strbyt (strinf stringbuffer) 0) $eof$)
%       (when (and (not (internp "BEGIN1")) (wneq filedescriptor 59))
%                                 %dont ruin stdin and Reduce
%        (setq stringbuffer "")
%        (conterror 94 "Read past EOF" (sysreadrec filedescriptor
%                        stringbuffer)))
         0
 )  )))) 
 
%------------------------------------------------------------------------------
% Write StringLength+1 characters from StringToWrite from the first position.
%------------------------------------------------------------------------------
 
(de syswriterec (filedescriptor stringtowrite stringlength)
     (cos-writer (cosstring stringtowrite) filedescriptor
                  (wplus2 stringlength 1))
   % (if (weq (cos-ckerr 17) 0 ) stringlength
   %    (conterror 93 " Write Error" (syswriterec filedescriptor
   %                                  stringtowrite stringlength)
     stringlength )  % ) )
 
%------------------------------------------------------------------------------
%      Close FileDescriptor, allowing it to be reused.
%------------------------------------------------------------------------------
 
(de sysclose (filedescriptor)
     (cos-close filedescriptor))
 
%------------------------------------------------------------------------------
%      Return the maximum buffer size.
%------------------------------------------------------------------------------
 
(de sysmaxbuffer (filedesc)  131) % most frontends like to cut a line
                                  % this is useful for a lineprinter
 
%------------------------------------------------------------------------------
% BinaryWrite and BinaryWriteBlock and BinaryopenWrite for fasl in / out.
%------------------------------------------------------------------------------
 
(de binaryopenwrite (filename)
     (sysopenwrite 0 filename))
 
(fluid '(onewordbuffer))
 
(de binarywrite (filedesc wordtowrite)
    (progn (setf (wgetv onewordbuffer 0) wordtowrite)
         (binarywriteblock filedesc (loc (wgetv onewordbuffer 0)) 1)
)   )
 
(de binarywriteblock (filedesc blockbase blocksize)
     (prog (len retval)
            (setq retval (cos-bwrit blocksize blockbase filedesc))
            (if (weq retval 0) (return len)
              (conterror  92 "Write Error" (BinaryWriteBlock
                            filedesc blockbase blocksize))
)     )     )
 
(de binaryopenread (filename)
     (sysopenread 0 filename))
 
(de binaryread (channel)    % Fortran's Channel-Number
                            % read 1 Word of 64 bits
      (if (weq (cos-bread 1 (loc (wgetv onewordbuffer 0)) channel)
                0)
           (wgetv onewordbuffer 0)
           (conterror 92 "Read Error" (BinaryRead channel))
)   )
 
(de binaryreadblock (channel blockbase blocksize)
      (cos-bread blocksize blockbase channel))
 
