%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for i860
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 5-Mar-90  (Winfried Neun)
%  added and refined (x-)code-address-to-symbol
% 10-Mar-89 (Winfried Neun)
%  Handling for i860 (SUN4). Complete redesign.
%  Also installed fluid *bruch* to allow gc (etc) to turn off
%  cntrl-c interrupt detection, which destroy system.
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is received.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(build-trap-message) 'internalfunction))

(fluid '(*bruch* i860-sigcp i860-sigaddr i860-sigcode i860-signumber
         *dump i860-sigpc i860-dumparray*))

% this is the interface for toploop

(de initializeinterrupts ()

	  % all signals are initiated expect for those in the following list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 18 19 20 21 22 24 25 26 28))) (sigsetup i)))))

% this installs the handler

(de sigsetup (signumber)
    (psl_sigset signumber (wgetv symfnc (id2int 'signalhandler))))

% this is the 'handler'

(setq i860-dumparray* (gtwarray 65))

(setq *bruch* 0)

(de signal-done ()
     (prog (handel)
	 (copywarray i860-dumparray* i860-sigcp 64)
         (setq handel (wgetv symfnc (id2int 'lisprpv)))
         (wputv i860-sigcp 4 handel)
)    )            %        'Brutal' version like Cray X-MP for UNICOS

(de lisprpv()
   (prog (mess)
    (setq *bruch* 17)
    (smalldump)
    (setq mess (atsoc i860-signumber '((1 . "Hangup")
                          (2 . "Interrupt")
                          (3 . "Quit")
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap")
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  i860-signumber)))
    (build-trap-message mess i860-sigpc)
    (setq i860-sigaddr 0 i860-sigcp 0) % for gc
    (return (stderror))))

(de smalldump()
  (when *dump
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn i860-signumber)
    (console-print-string " , signal code: 16#")
    (unixputn i860-sigcode)
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn i860-sigpc)
    (console-print-string " , signal stack: 16#")
    (unixputn i860-sigaddr)
    (console-print-string " , context: 16#")
    (unixputn i860-sigcp)
    (console-newline)
    (setq i860-sigaddr 0 i860-sigcp 0) % for gc
    (console-print-string "*** Registers  0-7 : ")
    (unixputn (wgetv i860-dumparray* 30)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 31)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 32)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 33)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 34)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 35)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 36)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 37)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Registers  8-15: ")
    (unixputn (wgetv i860-dumparray* 38)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 39)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 40)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 41)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 42)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 43)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 44)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 45)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Registers 16-23: ")
    (unixputn (wgetv i860-dumparray* 46)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 47)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 48)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 49)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 50)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 51)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 52)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 53)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Registers 24-31: ")
    (unixputn (wgetv i860-dumparray* 54)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 55)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 56)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 57)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 58)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 59)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 60)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 61)) (console-print-string " ")
     (console-newline)
    (console-print-string "*** Sigcontext     : ")
    (unixputn (wgetv i860-dumparray* 0)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 1)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 2)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 3)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 4)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 5)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 6)) (console-print-string " ")
    (unixputn (wgetv i860-dumparray* 7)) (console-print-string " ")
     (console-newline)
     (console-newline)
))

(compiletime (flag '(x-code-address-to-symbol) 'internalfunction))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in "
                                 (x-code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(fluid '(code-address* closest-address* closest-symbol*))

(de x-code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0))
        (mapobl #'(lambda (symbol)
                 (when (fcodep symbol)
                       (let ((address (inf (getfcodepointer symbol))))
                            (when (and (ileq address code-address*)
                                       (igreaterp address closest-address*))
                                  (setq closest-address* address)
                                  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de code-address-to-symbol (ad) (x-code-address-to-symbol ad))
