%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC/MULS32.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for HP RISC processor
% Author:       H. Melenk
% Created:      11 Sept  1989
% Modified:
% Mode:         Lisp
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Double length multiply and divide for the  processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
%
% Full 32 bit version : 28-June 1993 (H. Melenk)

(compiletime (flag '(wquotientdouble*) 'internalfunction))

(fluid '(*second-value* carry* doublearray*))
 
%----------------- unsigned add and subtract operations -------------------

% add and set carry
%
%   (a + b) -> (carry*,result)

(bothtimes (progn

(put 'addAndSetCarry 'opencode
       '(
         (ADD (reg 2)(reg 1)(reg 1))
           % move cf to carry*
         (ADDC (reg 0)(reg 0)(reg t1))
         (*MOVE (reg t1) ($FLUID carry*))
       ))

(put 'addAndSetCarry 'destroys '((reg 1)))

% add with carry
%
%   (a + b + carry*) -> (carry*,result)

(put 'addwithcarry 'opencode
       '(
           % move carry* to register CF
         (*MOVE ($fluid carry*)(reg t1))
         (*wshift (reg t1) 31)
         (ADD (reg t1)(reg t1)(reg t1))         
           % add with carry
         (ADDC (reg 2)(reg 1)(reg 1))
           % move cf to carry*
         (ADDC (reg 0)(reg 0)(reg t1))
         (*MOVE (reg t1) ($FLUID carry*))
       ))

(put 'addwithcarry 'destroys '((reg 1)))


% add and add carry
%
%   (a + b) -> result,  (*second-value* + carry) -> *second-value*

(put 'addAndAddCarry 'opencode
       '(
         (*move ($FLUID *second-value*) (reg t1))
           % add (reg 2) to (reg 1)
         (ADD (reg 2)(reg 1)(reg 1))
           % add carry to reg 3
         (ADDC (reg 0)(reg t1)(reg t1))
         (*move (reg t1) ($FLUID *second-value*))
       ))

(put 'addAndAddCarry 'destroys '((reg 1)))

% subtract with borrow
%
%  (a - (b + carry!*)) -> result, carry*=1 if borrow

(put 'subtractwithborrow 'opencode
       '(
           % move carry* to cf
         (*MOVE ($fluid carry*)(reg t1))
         (*WXOR (reg t1)(wconst 1))
         (*WSHIFT (reg t1) 31)
         (ADD (reg t1)(reg t1)(reg t1))    
           % subtract with borrow
         (SUBB (reg 1)(reg 2)(reg 1))
           % move new borrow to carry*
         (ADDC (reg 0)(reg 0)(reg t1))
         (*WXOR (reg t1)(wconst 1))
         (*MOVE (reg t1) ($FLUID carry*))
       ))

(put 'subtractwithborrow  'destroys '((reg 1)))

%------------------- unsigned greaterp ---------------------

% code based on integer carry; 
 

(put 'ugreaterp* 'opencode 
   % returns 1 if arg1 > arg2 unsigned.
'( 
   (sub (reg 2)(reg 1)(reg 0))        % compare, setting carry if r1>r2
   (addc (reg 0)(reg 0)(reg 1))       % move carry to reg 1
   (*wxor (reg 1)(wconst 1))
))

(put 'ugreaterp*  'destroys '((reg 1)))

(ds ugreaterp(a b)(eq 1 (ugreaterp* a b)))

)) % bothtimes

%-------------------------- multiply, divide -------------------------

(setq doublearray* (wand (gtwarray 4) -8))

(de wxtimes2 (x y)
(SETQ BOESEWICHT 2)
	(wddtimes doublearray* x y)
        (setq *second-value* (wgetv doublearray* 0))
	(wgetv doublearray* 1))

(bothtimes(progn

(put 'dummy-local 'opencode '((*move (reg 1)(reg 1))))

(ds wtimesdouble (x y)
        (dummy-local) % force values to frame
        (wdtimes-local (loc x) (loc y) doublearray*))

(put ' wdtimes-local 'opencode '(
  (fullword 16#2720100A)          % FLDWS   0(%arg1),%fr10
  (fullword 16#2740100B)          % FLDWS   0(%arg0),%fr11
  (fullword 16#394B470C)          % XMPYU 10,11,12
  (fullword 16#2F00120C)          % FSTDS   12,0(%arg2)
  (*move (memory (reg 3) (wconst 0)) (reg 2))
  (*move (reg 2)($fluid *second-value*))
  (*move (memory (reg 3) (wconst 4)) (reg 1))
))
))

(flag '(wtimesdouble) 'lose)


(commentoutcode
(de wquotientdouble(a0 a1 b)
   (prog(q rem)
       % unsigned double quotient using the routine
       % for signed quotient.
       % simple case: neither a nor b formally negative
     (when (and (eq b 1)(eq a0 0))(setq *second-value* 0)(return a1)) 
     (when (and (wgeq a0 0) (wgreaterp b 0))
           (return (wquotientdouble* a0 a1 b)))
       % approximate quotient
     (setq q (wquotientdouble*
              (wshift a0 -1)
              (wor (wshift a0 (plus bitsperword -1)) (wshift a1 -1))
              (wshift b -1)))
     (setq rem
            (wor (wshift *second-value* 1)
                 (wand a1 1)))
     (when (eq 0 (wand b 1)) (go ret))
   % last bit of b is 1: adjust remainder
     (setq carry* 0)
     (setq rem (subtractwithborrow rem q))
     (when (weq carry* 0) (go ret))
   % if carry, q has been to big
   % decrease q until rem swaps again
     (setq carry* 0)
     (while (weq carry* 0)
       (setq q (isub1 q))
       (setq rem (addAndSetCarry b rem)))
   ret
     (setq *second-value* rem)
     (return q) ))
)

(de wquotientdouble(a0 a1 b)
   (prog(q rem a00 a11 rcarry)
(SETQ BOESEWICHT 4)
       % unsigned double quotient using the routine
       % for signed quotient.
       % simple case: neither a nor b formally negative
     (when (and (eq b 1)(eq a0 0))(setq *second-value* 0)(return a1))
     (when (and (wgeq a0 0) (wgreaterp b 0))
           (return (wquotientdouble* a0 a1 b)))
     (when (wlessp a0 0)(go case-2))

       % case 1: a0 is shorter than b
       % approximate quotient by ((a0,a1)/2) : (b/2)
     (setq q (wquotientdouble*
              (wshift a0 -1)
              (wor (wshift a0 (plus bitsperword -1)) (wshift a1 -1))
              (wshift b -1)))
     (setq rem
            (wor (wshift *second-value* 1)
                 (wand a1 1)))
     (when (eq 0 (wand b 1)) (go ret))
     (go korr)

 case-2
     % case 2: a0 is not shorter than b
     % approximate quotient by ((a0,a1)/4) : (b/2)

     (setq q (wquotientdouble*
              (wshift a0 -2)
              (wor (wshift a0 (plus bitsperword -2)) (wshift a1 -2))
              (wshift b -1)))

        % adjust quotient
     (cond((not (eq 0 (wshift q (wdifference 1 bitsperword))))
           (setq q -1)(go korr))  % biggest possible value if overflow
          (t (setq q (wshift q 1))))

        % adjust remainder
     (setq rem
            (wor (wshift *second-value* 2)
                 (wand a1 3)))
   % (when (and (eq 0 (wand b 1)) (eq 0 (wand a1 3))) (go ret))

   % adjust quotient, recompute remainder
  korr 
     (setq a11 (wtimesdouble q b))
     (setq carry* 0)
     (setq rem (subtractwithborrow a1 a11))
     (setq a00 (subtractwithborrow a0 *second-value*))
         % underflow or overflow?
     (cond ((not (eq carry* 0)) (setq q (isub1 q)) (go korr))
           ((or (not (eq a00 0)) (ugreaterp rem b) (eq rem b))
            (setq q (iadd1 q)) (go korr)))
   ret
        %  (setq a11 (wtimesdouble q b))
        %  (setq a00 *second-value*)
        %  (setq carry* 0)
        %  (setq a11 (addwithcarry a11 rem))
        %  (setq a00 (addwithcarry a00 0))
        %  (when (or (not (eq a1 a11))(not (eq a0 a00)) 
        %            (eq rem b)(ugreaterp rem b))
        %        (set 'a a0) (set 'b a1)(set 'c b)
        %        (print (list "mist" (hilfe q)(hilfe rem)))
        %        (yesp "catastrophic divide error ++++")
        %        (error 99 "FALSCH"))
     (setq *second-value* rem)
     (return q) ))


(de hilfe(u)
  (let((v (gtpos 1)))
       (iputv v 1 u)
       v))


(lap '((*entry wquotientdouble* expr 3)
      % called with a double length number in params 1 and 2
      % and a single length number in par 3.
      % Result is the single length quotient.
      % the remainder is placed in a fluid variable.

      % Code is taken from the Assemble Language Ref Man, p 5.6
      % where one can find some comments too.


	(*alloc 0)

	(movb (compl >= )  (reg 1) (reg 5) (label check_mag))
	(add	(reg r0) (reg 2) (reg 4))
	(sub	(reg r0) (reg 4) (reg 4))
	(subb	(reg r0) (reg 5) (reg 5))
  check_mag
	(subt (compl = ) (reg r0) (reg 3) (reg t1))

	(ds	(reg r0) (reg t1) (reg r0))
	(add	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 1
	(addc	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 2
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 3
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 4
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 5
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 6
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 7
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 8
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 9
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 10
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 11
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 12
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 13
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 14
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 15
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 16
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 17
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 18
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 19  
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 20
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 21
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 22
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 23
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 24
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 25
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 26
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 27
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 28
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 29
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 30
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 31
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 32
	(addc	(reg 4) (reg 4) (reg 4))
	(addbf (compl < ) (reg 5) (reg r0) (label finish))
        (nop)
	(add (compl < ) (reg 3) (reg r0) (reg r0))
	(add (compl tr) (reg 5) (reg 3) (reg 5))
	(addl	(reg 5) (reg t1) (reg 5))
    finish
	(add (compl >= ) (reg 1) (reg r0) (reg r0))
	(sub	(reg r0) (reg 5) (reg 5))
	(xor (compl >= ) (reg 1) (reg 3) (reg r0))
	(sub	(reg r0) (reg 4) (reg 4))
	(*move	(reg 4) (reg 1))
	(*move	(reg 5) ($fluid *second-value*))
	(*exit 0)))


(flag '(wquotientdouble) 'lose)

(lap '((*entry wdivide expr 2)
        (*alloc 0)
	(*move (reg 2) (reg 3))
	(*move (reg 1) (reg 2))
	(extrs (reg 2) 0 1 (reg 1))
	(movb (compl >= )  (reg 1) (reg 5) (label check_mag))
	(add	(reg r0) (reg 2) (reg 4))
	(sub	(reg r0) (reg 4) (reg 4))
	(subb	(reg r0) (reg 5) (reg 5))
  check_mag
	(subt (compl = ) (reg r0) (reg 3) (reg t1))

	(ds	(reg r0) (reg t1) (reg r0))
	(add	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 1
	(addc	(reg 4) (reg 4) (reg 4))
	(ds	(reg 5) (reg 3) (reg 5))  % 2
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 3
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 4
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 5
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 6
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 7
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 8
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 9
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 10
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 11
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 12
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 13
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 14
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 15
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 16
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 17
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 18
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 19  
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 20
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 21
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 22
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 23
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 24
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 25
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 26
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 27
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 28
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 29
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 30
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 31
        (addc   (reg 4) (reg 4) (reg 4))
        (ds     (reg 5) (reg 3) (reg 5))  % 32
	(addc	(reg 4) (reg 4) (reg 4))
	(addbf (compl < ) (reg 5) (reg r0) (label finish))
        (nop)
	(add (compl < ) (reg 3) (reg r0) (reg r0))
	(add (compl tr) (reg 5) (reg 3) (reg 5))
	(addl	(reg 5) (reg t1) (reg 5))
    finish
	(add (compl >= ) (reg 1) (reg r0) (reg r0))
	(sub	(reg r0) (reg 5) (reg 5))
	(xor (compl >= ) (reg 1) (reg 3) (reg r0))
	(sub	(reg r0) (reg 4) (reg 4))
	(*move	(reg 4) (reg 1))
	(*move	(reg 5) ($fluid *second-value*))
	(*exit 0)))

